import timeit
import logging
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError, Timeout
import json
from nuaal.definitions import DATA_PATH
from nuaal.utils import get_logger

class RestBase(object):
    """
    Parent class for all REST-like connections
    """
    def __init__(self, url, username=None, password=None, api_base_path=None, verify_ssl=False, DEBUG=False, con_type=None):
        """

        :param url: URL or IP address of the target machine
        :param username: Username for authentication
        :param password: Password for authentication
        :param api_base_path: Base path for the API resource, such as "/api/v1"
        :param verify_ssl: Enable SSL certificate verification. For self-signed certificates, set this to False
        :param DEBUG: Enable debugging output
        :param con_type: String representation of connection type, set by child classes
        """
        # Disable logging for external libraries
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)

        self.url = url
        self.logger = get_logger(name="{}-{}".format(con_type if con_type else 'REST', self.url), DEBUG=DEBUG)
        self.username = username
        self.password = password
        self.api_base_path = api_base_path
        self.verify_ssl = verify_ssl
        self.DEBUG = DEBUG
        self.path_base = self.url + self.api_base_path
        self.headers = {"Content-type": "application/json"}
        self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}

    def __enter__(self):
        """Allow usage of Python Context Manager"""
        self._initialize()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Allow usage of Python Context Manager"""
        pass

    def _initialize(self):
        """
        Function for credentials loading and session preparation. Might be overwritten in child classes

        :return: ``None``
        """
        if self.verify_ssl == False:
            # Disable certificate warning
            try:
                requests.packages.urllib3.disable_warnings()
            except:
                self.logger.warning(msg="Failed to disable Certificate Warnings")
        if self.username is None or self.password is None:
            self.logger.info(msg="No credentials provided, using cached instead.")
            if self._load_credentials():
                self.logger.info(msg="Credentials successfully loaded.")
            else:
                self.logger.info(msg="Failed to load valid credentials.")
        else:
            self._authorize()

    def _authorize(self):
        """
        Connection-specific function for handling authorization. Overridden by child classes

        :return: ``None``
        """
        pass

    def _get(self, path, params=None):
        """
        Wrapper function for GET method of the requests library

        :param path: Path of the API resource used in URL
        :param params: Parameters for the request
        :return: Instance of ``requests`` response object
        """
        response = None
        try:
            self.logger.debug(msg="_GET: Path: '{}', Params: '{}'".format(path, params))
            if params:
                response = requests.get(url=self.path_base + path, **self.common_headers, params=params)
            else:
                response = requests.get(url=self.path_base + path, **self.common_headers)
        except ConnectionError:
            self.logger.critical(msg="_GET: Could not connect to {}. Wrong address?".format(self.url))
        except Exception as e:
            self.logger.error(msg="_GET: Encountered unhandled Exception: {}".format(repr(e)))
        finally:
            return response

    def _post(self, path, data=None, files=None, params=None):
        """
        Wrapper function for POST method of the requests library

        :param path: Path of the API resource used in URL
        :param data: Data payload for POST request. JSON string
        :param dict files: Dictionary with files to upload
        :param params: Parameters for the request
        :return: Instance of ``requests`` response object
        """
        response = None
        args = {}
        if data:
            args["data"] = data
        if files:
            args["files"] = files
            del self.common_headers["headers"]["Content-type"]
        if params:
            args["params"] = params
        try:
            self.logger.debug(msg="_POST: Path: '{}', Data: '{}', Params: '{}'".format(path, data, params))
            response = requests.post(url=self.path_base + path, **args, **self.common_headers)
        except ConnectionError:
            self.logger.critical(msg="_POST: Could not connect to {}. Wrong address?".format(self.url))
        except Exception as e:
            self.logger.error(msg="_POST: Encountered unhandled Exception: {}".format(repr(e)))
        finally:
            self.common_headers["headers"]["Content-type"] = "application/json"
            return response

    def _delete(self, path, params):
        """
        Wrapper function for DELETE method of the requests library

        :param path: Path of the API resource used in URL
        :param params: Parameters for the request
        :return: Instance of ``requests`` response object
        """
        response = None
        try:
            self.logger.debug(msg="_DELETE: Path: '{}', Params: '{}'".format(path, params))
            if params:
                response = requests.delete(url=self.path_base + path, **self.common_headers, params=params)
            else:
                response = requests.delete(url=self.path_base + path, **self.common_headers)
        except ConnectionError:
            self.logger.critical(msg="_DELETE: Could not connect to {}. Wrong address?".format(self.url))
        except Exception as e:
            self.logger.error(msg="_DELETE: Encountered unhandled Exception: {}".format(repr(e)))
        finally:
            return response

    def _put(self, path, data=None, files=None, params=None):
        """
        Wrapper function for PUT method of the requests library

        :param path: Path of the API resource used in URL
        :param data: Data payload for POST request. JSON string
        :param params: Parameters for the request
        :return: Instance of ``requests`` response object
        """
        response = None
        args = {}
        if data:
            args["data"] = data
        if files:
            args["files"] = files
            del self.common_headers["headers"]["Content-type"]
        if params:
            args["params"] = params
        try:
            self.logger.debug(msg="_PUT: Path: '{}', Data: '{}', Params: '{}'".format(path, data, params))
            response = requests.put(url=self.path_base + path, **args, **self.common_headers)
        except ConnectionError:
            self.logger.critical(msg="_PUT: Could not connect to {}. Wrong address?".format(self.url))
        except Exception as e:
            self.logger.error(msg="_PUT: Encountered unhandled Exception: {}".format(repr(e)))
        finally:
            self.common_headers["headers"]["Content-type"] = "application/json"
            return response

    def _response_handler(self, response):
        """
        Function for handling request's response objects. Main purpose of this action is to handle HTTP return codes.
        In case of JSON formatted reply, returns this data as dictionary. In case of errors, either string representation
        of data is returned, or None. Status code is returned alongside the content to allow further processing if needed.

        :param response: Instance of request's response object.
        :return: content, status code
        """
        if response is None:
            return None, None
        try:
            status_code = response.status_code
            if status_code in range(200, 208):
                self.logger.debug(msg="Response_Handler: Server returned Status Code: {} OK".format(status_code))
                return response.json(), status_code
            if status_code == 400:
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, Bad Request.".format(status_code))
                try:
                    return response.json(), status_code
                except Exception as e:
                    self.logger.error(msg="Response_Handler: Could not return JSON error representation. Raw Text: {} Exception: {}.".format(response.text, status_code))
                    return response.text, status_code
            if status_code == 401:
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, Unauthorized.".format(status_code))
                return response.json(), status_code
            if status_code == 403:
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, Forbidden.".format(status_code))
                return response.text, status_code
            if status_code == 404:
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, Not Found.".format(status_code))
                return response.text, status_code
            if status_code in range(400, 500):
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, cannot process response.".format(status_code))
                return None, status_code
            if status_code in range(500, 600):
                self.logger.error(msg="Response_Handler: Server returned Status Code: {}, cannot process response.".format(status_code))
                return None, status_code
        except Exception as e:
            self.logger.critical(msg="Response Handler: Unhandled exception occurred. Exception: {}".format(repr(e)))




import json
import datetime

class CiscoApiModel(object):
    def __init__(self):
        pass

    def str_to_datetime(self, date_str, date_format="%Y-%m-%d"):
        return datetime.datetime.strptime(date_str, date_format)

class EoxModel(CiscoApiModel):
    def __init__(self, eox_record):
        self.all_keys = ["EOLProductID", "ProductIDDescription", "EOXExternalAnnouncementDate", "EndOfSaleDate", "EndOfSWMaintenanceReleases", "EndOfSecurityVulSupportDate", "EndOfRoutineFailureAnalysisDate", "EndOfServiceContractRenewal", "EndOfServiceContractRenewal", "LastDateOfSupport", "EndOfSvcAttachDate", "UpdatedTimeStamp"]
        self.eox_record = eox_record if isinstance(eox_record, dict) else json.loads(eox_record)
        self.EOLProductID = self.eox_record["EOLProductID"]
        self.ProductIDDescription = self.eox_record["ProductIDDescription"]
        self.ProductBulletinNumber = self.eox_record["ProductBulletinNumber"]
        self.LinkToProductBulletinURL = self.eox_record["LinkToProductBulletinURL"]
        self.EOXExternalAnnouncementDate = self.str_to_datetime(self.eox_record["EOXExternalAnnouncementDate"]["value"])
        self.EndOfSaleDate = self.str_to_datetime(self.eox_record["EndOfSaleDate"]["value"])
        self.EndOfSWMaintenanceReleases = self.str_to_datetime(self.eox_record["EndOfSWMaintenanceReleases"]["value"])
        self.EndOfSecurityVulSupportDate = self.str_to_datetime(self.eox_record["EndOfSecurityVulSupportDate"]["value"])
        self.EndOfRoutineFailureAnalysisDate = self.str_to_datetime(self.eox_record["EndOfRoutineFailureAnalysisDate"]["value"])
        self.EndOfServiceContractRenewal = self.str_to_datetime(self.eox_record["EndOfServiceContractRenewal"]["value"])
        self.LastDateOfSupport = self.str_to_datetime(self.eox_record["LastDateOfSupport"]["value"])
        self.EndOfSvcAttachDate = self.str_to_datetime(self.eox_record["EndOfSvcAttachDate"]["value"])
        self.UpdatedTimeStamp = self.str_to_datetime(self.eox_record["UpdatedTimeStamp"]["value"])


if __name__ == '__main__':
    eox_dict = {
      "EOLProductID": "WS-C3560G-48PS-S",
      "ProductIDDescription": "Catalyst 3560 48 10/100/1000T PoE + 4 SFP + IPB Image",
      "ProductBulletinNumber": "EOL8044",
      "LinkToProductBulletinURL": "http://www.cisco.com/en/US/prod/collateral/switches/ps5718/ps5023/eol_c51-696372.html",
      "EOXExternalAnnouncementDate": {
        "value": "2012-01-31",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfSaleDate": {
        "value": "2013-01-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfSWMaintenanceReleases": {
        "value": "2014-01-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfSecurityVulSupportDate": {
        "value": "2016-01-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfRoutineFailureAnalysisDate": {
        "value": "2014-01-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfServiceContractRenewal": {
        "value": "2017-04-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "LastDateOfSupport": {
        "value": "2018-01-31",
        "dateFormat": "YYYY-MM-DD"
      },
      "EndOfSvcAttachDate": {
        "value": "2014-01-30",
        "dateFormat": "YYYY-MM-DD"
      },
      "UpdatedTimeStamp": {
        "value": "2013-09-18",
        "dateFormat": "YYYY-MM-DD"
      },
      "EOXMigrationDetails": {
        "PIDActiveFlag": "Y",
        "MigrationInformation": "Catalyst 3560X 48 Port PoE IP Base",
        "MigrationOption": "Enter PID(s)",
        "MigrationProductId": "WS-C3560X-48P-S",
        "MigrationProductName": "",
        "MigrationStrategy": "",
        "MigrationProductInfoURL": "http://www.cisco.com/en/US/products/ps10745/index.html and http://www.cisco.com/en/US/products/ps10744/index.html"
      },
      "EOXInputType": "ShowEOXByPids",
      "EOXInputValue": "WS-C3560G-48PS-S "
    }

    eox_entry = EoxModel(eox_dict)
    print(eox_entry.EOXExternalAnnouncementDate)


import os
import json
from nuaal.utils import get_logger

class Reader:
    def __init__(self):
        self.logger = get_logger("Reader")
        raise NotImplemented("This class currently being tested.")

    def filename_check(self, filename):
        if os.path.exists(filename):
            self.logger.debug(msg="Filename {} exists.".format(filename))
            print(os.path.dirname(__file__))
            return filename
        else:
            cwd = os.path.dirname(__file__)
            filepath = os.path.join(cwd, filename)
            if os.path.exists(filepath):
                self.logger.debug(msg="Filename {} exists.".format(filename))
                return os.path.abspath(filepath)
            else:
                self.logger.error(msg="Filename {} does not exist.".format(filename))
                return None

    def load_json(self, filename):
        filename = self.filename_check(filename)
        data = None
        if filename:
            with open(filename, mode='r') as f:
                try:
                    data = json.load(f)
                except Exception as e:
                    self.logger.critical(msg="Failed to load JSON file, Exception: {}".format(repr(e)))
                finally:
                    return data
        else:
            self.logger.critical(msg="Could not open file {}, file does not exist".format(filename))
            return data

    def load_csv(self, filename, delimiter=","):
        filename = self.filename_check(filename)
        print(filename)
        lines = None
        if filename:
            with open(filename, mode='r') as f:
                try:
                    temp_lines = [x.split(delimiter) for x in f.readlines()]
                    lines = []
                    for line in temp_lines:
                        lines.append([x.strip() for x in line])
                    return lines
                except Exception as e:
                    self.logger.critical(msg="Failed to load CSV file, Exception: {}".format(repr(e)))
                    return lines

        else:
            self.logger.critical(msg="Could not open file {}, file does not exist".format(filename))
            return lines




from nuaal.utils import get_logger
from nuaal.definitions import DATA_PATH
from nuaal.utils import int_name_convert, mac_addr_convert, interface_split, vlan_range_expander, vlan_range_shortener
from nuaal.utils import Filter


class BaseModel(object):
    """
    This is a parent object that all high-level API models inherit from.
    """
    def __init__(self, name=None, DEBUG=False):
        """

        :param name:
        :param DEBUG:
        """
        self.model_name = name if name else "BaseModel"
        self.DEBUG = DEBUG
        self.logger = get_logger(name="{}-Model".format(self.model_name), DEBUG=self.DEBUG)

    def __str__(self):
        return "{}-Model".format(self.model_name)

    def __repr__(self):
        return "{}-Model".format(self.model_name)

class DeviceBaseModel(BaseModel):
    """
    This is a parent object for device models.
    """
    def __init__(self, name="DeviceBaseModel", DEBUG=False):
        """

        :param name:
        :param DEBUG:
        """
        super(DeviceBaseModel, self).__init__(name=name, DEBUG=DEBUG)
        self.interfaces = {}
        self.vlans = {}
        self.inventory = []
        self.device_info = {
            "mgmtIpAddress": u"",
            "hostname": u"",
            "vendor": u"",
            "platform": u"",
            "swVersion": u"",
            "uptime": u""
        }
        self.interface_model = {
            "description": None,
            "interfaceType": None,
            "className": None,
            "status": None,
            "macAddress": None,
            "adminStatus": None,
            "speed": None,
            "portName": None,
            "untaggedVlanId": None,
            "taggedVlanIds": [],
            "duplex": None,
            "portMode": None,
            "portType": None,
            "ipv4Mask": None,
            "ipv4Address": None,
            "mediaType": None,
        }
        self.vlan_model = {
            "vlanId": "",
            "name": "",
            "status": "",
            "untaggedPorts": [],
            "taggedPorts": []
        }
        self.inventory_model = {
            "name": "",
            "description": "",
            "partNumber": "",
            "serialNumber": "",
            "version": ""
        }

    def get_interfaces(self):
        """

        :return:
        """
        raise NotImplemented("Function get_interfaces has not yet been implemented for model {}".format(self.model_name))

    def get_vlans(self):
        """

        :return:
        """
        raise NotImplemented("Function get_vlans has not yet been implemented for model {}".format(self.model_name))

    def get_neighbors(self):
        """

        :return:
        """
        raise NotImplemented("Function get_neighbors has not yet been implemented for model {}".format(self.model_name))


from nuaal.utils import *
from nuaal.definitions import DATA_PATH
from nuaal.Parsers import Patterns
import json
import re
import timeit


class ParserModule(object):
    """
    This class provides necessary functions for parsing plaintext output of network devices. Uses patterns from ``PatternsLib`` for specified device type.
    The outputs are usually lists of dictionaries, which contain keys based on name groups of used regex patterns.
    """
    def __init__(self, device_type, DEBUG=False):
        """

        :param str device_type: String representation of device type, such as `cisco_ios`
        :param bool DEBUG: Enables/disables debugging output
        """
        raise DeprecationWarning("This Class is deprecated")
        self.device_type = device_type
        self.DEBUG = DEBUG
        self.logger = get_logger(name="ParserModule-{}".format(device_type), DEBUG=DEBUG)
        self.logger.info(msg="Creating ParserModule Object for {}".format(device_type))
        self.patterns = Patterns(device_type=device_type).get_patterns()

    def match_single_pattern(self, text, pattern):
        """
        This function tries to match given regex ``pattern`` against given ``text``. If ``pattern`` contains named groups,
        list of dictionaries with these groups as keys is returned. If ``pattern`` does not contain any named groups,
        list of matching strings is returned.

        :param str text:
        :param pattern: ``re`` compiled regex pattern
        :return: List of matches, either dictionaries or strings
        """
        if not re.search(pattern=pattern, string=text):
            return []
        named_groups = [x for x in pattern.groupindex.keys() if isinstance(x, str)]
        if len(named_groups) == 0:
            return re.findall(pattern=pattern, string=text)
        else:
            entries = []
            for m in re.finditer(pattern=pattern, string=text):
                entry = {}
                for group_name in named_groups:
                    try:
                        entry[group_name] = int(m.group(group_name))
                    except (ValueError, TypeError):
                        entry[group_name] = m.group(group_name)
                    except IndexError:
                        entry[group_name] = None
                entries.append(entry)
            return entries

    def update_entry(self, orig_entry, new_entry):
        """
        This function simply updates dictionary ``orig_entry`` with keys and values from ``new_entry``. Only ``None`` values in ``orig_entry``
        are updated. Keys with value ``None`` from ``new_entry`` are also used in order to ensure coherent output.

        :param dict orig_entry: Original dictionary to be updated based on entries in ``new_entry``
        :param dict new_entry: New dictionary containing new data which should be added to ``orig_entry``
        :return: Updated dictionary
        """
        if not isinstance(orig_entry, dict) or not isinstance(new_entry, dict):
            raise TypeError()
        for k, v in new_entry.items():
            try:
                if orig_entry[k] is None:
                    orig_entry[k] = v
            except KeyError:
                orig_entry[k] = None
        self.logger.debug(msg="Updated entry {} based on {}".format(orig_entry, new_entry))
        return orig_entry

    def match_multi_pattern(self, text, patterns):
        """
        This functions tries to match multiple regex ``patterns`` against given ``text`` .

        :param str text: Text output to be processed
        :param lst patterns: List of ``re`` compiled patterns for parsing.
        :return: Dictionary with names of all groups from *all* ``patterns`` as keys, with matching strings as values.
        """
        named_groups = []
        entry = {}
        match_counter = 0
        start_time = timeit.default_timer()
        # Populate named_groups
        for pattern in patterns:
            for group_name in pattern.groupindex.keys():
                if isinstance(group_name, str) and group_name not in named_groups:
                    named_groups.append(group_name)
                    entry[group_name] = None
        self.logger.debug(msg="Found {} groups in patterns: {}".format(len(named_groups), named_groups))
        for pattern in patterns:
            match = re.search(pattern=pattern, string=text)
            if match:
                match_counter += 1
                pattern_match = {}
                for group_name in [x for x in pattern.groupindex.keys() if isinstance(x, str)]:
                    try:
                        pattern_match[group_name] = int(match.group(group_name))
                    except (TypeError, ValueError):
                        pattern_match[group_name] = match.group(group_name)
                entry = self.update_entry(entry, pattern_match)
                if None not in list(entry.values()):
                    break
        self.logger.debug(msg="MultiPatternMatch: Matched {} pattern(s) in {} miliseconds.".format(match_counter, (timeit.default_timer() - start_time)*1000))
        return entry

    def _level_zero(self, text, patterns):
        """
        This function handles parsing of less complex plaintext outputs, which can be parsed in  one step.

        :param str text: Plaintex output which will be parsed
        :param patterns: List of compiled regex patterns, which are used to parse the ``text``
        :return: List dics (if ``patterns`` contain named groups) or list of strings (if they don't)
        """
        if not isinstance(text, str):
            self.logger.debug(msg="Level Zero: Expected string, got {}".format(type(text)))
            return []
        level_zero_outputs = []
        if not isinstance(patterns, list):
            patterns = list(patterns)
        for pattern in patterns:
            output = self.match_single_pattern(text=text, pattern=pattern)
            if not output:
                continue
            if len(output) == 0:
                continue
            else:
                level_zero_outputs = output
                break
        if len(level_zero_outputs) == 0:
            self.logger.warning(msg="Level Zero found no matches.")
        elif isinstance(level_zero_outputs[0], str):
            self.logger.debug(msg="Level Zero: Found {} matches without named groups.".format(len(level_zero_outputs)))
        elif isinstance(level_zero_outputs[0], dict):
            self.logger.debug(msg="Level Zero: Found {} matches without named groups.".format(len(level_zero_outputs)))
        else:
            self.logger.critical(msg="Level Zero: Unexpected event when trying to match {} with patterns {}.".format(text, patterns))
        return level_zero_outputs

    def _level_one(self, text, command):
        """
        This function handles parsing of more complex plaintext outputs. First, the output of ``_level_zero()`` function is retrieved and then further parsed.

        :param str text: Plaintex output of given ``command``, which will be parsed
        :param str command: Command string used to generate the output
        :return: List of dictionaries
        """
        level_one_outputs = []
        level_zero_outputs = self._level_zero(text=text, patterns=self.patterns["level0"][command])
        if len(level_zero_outputs) == 0:
            self.logger.error(msg="Level Zero returned 0 outputs for command {}".format(command))
            return level_zero_outputs
        if command not in self.patterns["level1"].keys():
            self.logger.warning(msg="Command {} is not a 'level1' command.".format(command))
            return level_zero_outputs
        self.logger.debug(msg="Level Zero returned {} outputs.".format(len(level_zero_outputs)))
        if isinstance(level_zero_outputs[0], dict):
            for level_zero_entry in level_zero_outputs:
                entry = level_zero_entry
                for key, patterns in self.patterns["level1"][command].items():
                    try:
                        entry[key] = self._level_zero(text=level_zero_entry[key], patterns=patterns)
                    except KeyError:
                        self.logger.error(msg="Level Zero did not return entry with key {}.".format(key))
                        entry[key] = None
                    except TypeError as e:
                        self.logger.error(msg="Level Zero returned {} for key {}.".format(type(level_zero_entry[key]), key))
                        entry[key] = None
                level_one_outputs.append(entry)
        elif isinstance(level_zero_outputs[0], str):
            all_patterns = []
            for key, patterns in self.patterns["level1"][command].items():
                all_patterns += patterns
            for level_zero_entry in level_zero_outputs:
                entry = {}
                entry = self.match_multi_pattern(text=level_zero_entry, patterns=all_patterns)
                level_one_outputs.append(entry)
        return level_one_outputs

    def command_mapping(self, command):
        """
        This function determines the max_level of command - based on level of complexity of the output, the parsing is processed in 1 or 2 steps.
        Used for :ref:`autoparse <autoparse>`

        :param str command: Command used to generate the output, eg. `show vlan brief`
        :return: (str) Highest level of specified command (top PatternsLib Key)
        """
        levels = ["level0", "level1"]
        max_level = None
        for level in levels:
            if command in list(self.patterns[level].keys()):
                max_level = level
        self.logger.debug(msg="Command '{}' level is: {}".format(command, level))
        return max_level

    def autoparse(self, text, command):
        """
        .. _autoparse:

        The main entry point for parsing, given just the ``text`` output and ``command`` it determines the proper way to parse the output and returns result.

        :param str text: Text output to be processed
        :param str command: Command used to generate ``text`` output. Based on this parameter, correct regex patterns are selected.
        :return: List of found entities, usually list of dictionaries
        """
        command_level = self.command_mapping(command=command)
        parsed_output = None
        if command_level == "level0":
            return self._level_zero(text=text, patterns=self.patterns["level0"][command])
        elif command_level == "level1":
            return self._level_one(text=text, command=command)
        else:
            self.logger.critical(msg="AutoParse: Unknown level for command: '{}'".format(command))



from nuaal.connections.api.ciscoapi import CiscoApiBase
import json


class CiscoPsirtApi(CiscoApiBase):
    def __init__(self, username=None, password=None, api_user_id=None, DEBUG=False):
        super(CiscoPsirtApi, self).__init__(username=username,
                                            password=password,
                                            api_user_id=api_user_id,
                                            api_base_path="/security/advisories",
                                            con_type="CiscoEoxApi",
                                            DEBUG=DEBUG
                                            )

    def psirt_by_sw(self, sw_type, sw_version):
        response = self.get(path="/{}".format(sw_type), params={"version": sw_version})
        print(json.dumps(response, indent=2))

if __name__ == '__main__':
    psirt_client = CiscoPsirtApi(api_user_id="ad", DEBUG=True)
    psirt_client._initialize()
    psirt_client.psirt_by_sw(sw_type="ios", sw_version="15.2(1)E1")


from nuaal.connections.api.RestBase2 import RestBase2
from nuaal.definitions import DATA_PATH
from nuaal.utils import check_path
import json
import os
import copy
import requests
import re
import time


class DnacBase(RestBase2):

    def __init__(self, url, username=None, password=None, api_base_path="/api/v1", verify_ssl=True, DEBUG=False):
        """

        :param url: URL of the APIC-EM, such as 'https://sandboxapicem.cisco.com'
        :param username: Username for authentication
        :param password: Password for authentication
        :param api_base_path: Base path for version-1 API
        :param verify_ssl: Enable/disable verification of SSL Certificate
        :param DEBUG: Enable/disable debugging output
        """
        super(DnacBase, self).__init__(url=url, username=username, password=password,
                                       api_base_path=api_base_path, verify_ssl=verify_ssl,
                                       DEBUG=DEBUG, con_type="DNAC")
        self._initialize()

    def _authorize(self):
        if "X-Auth_Token" in self.headers.keys():
            del self.headers["X-Auth-Token"]
            self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
        response = None
        try:
            self.logger.debug(msg="Requesting new Auth Token.")
            response = requests.request(method="POST", url=self.url + "/api/system/v1/auth/token", auth=(self.username, self.password), **self.common_headers)
        except requests.exceptions.ConnectionError:
            self.logger.critical(msg="Could not connect to {}".format(self.url))
            self.authorized = False
            return False
        status_code = response.status_code
        if status_code == 401:
            self.logger.critical(msg="Could not get Auth Token. Reason: Invalid Credentials. Please provide valid credentials.")
            self.authorized = False
            return False
        if status_code == 200:
            response = response.json()
            if "Token" in response.keys():

                self.headers["X-Auth-Token"] = response["Token"]
                self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
                self.authorized = True
                self.logger.info(msg="Successfully updated Auth Token: {}".format(self.headers['X-Auth-Token']))
                self._store_credentials()
                return True

    def _load_credentials(self):
        """
        Function for loading credentials information stored in data/apicem_credentials.json file.

        :return: ``None``
        """
        credentials = self._load_data(path="dnac_credentials.json")
        for value in credentials.values():
            if value == "":
                return False
        self.username = credentials["username"]
        self.password = credentials["password"]
        self.headers["X-Auth-Token"] = credentials["serviceTicket"]
        self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
        self.authorized = True
        return True

    def _store_credentials(self):
        """
        Function for storing credentials information in data/apicem_credentials.json file. This function is called after successfully obtaining Service Ticket

        :return: ``None``
        """
        self.logger.info(msg="Storing credentials.")
        data = {"username": self.username, "password": self.password, "serviceTicket": self.headers["X-Auth-Token"]}
        self._store_data(data=data, path="dnac_credentials.json")


if __name__ == '__main__':
    dnac = DnacBase(url="https://sandboxdnac.cisco.com", verify_ssl=True, DEBUG=True)
    dnac._authorize()



import logging
from nuaal.connections.cli.CliBase import CliBaseConnection
from nuaal.connections.cli.Cisco_IOS_Cli import Cisco_IOS_Cli
from nuaal.connections.cli.CliMultiRunner import CliMultiRunner
from nuaal.connections.cli.GetCliHandler import GetCliHandler
# Disable error logging for Paramiko library
logging.getLogger("paramiko").setLevel(logging.CRITICAL)



import os
import sys

USER_DIR = os.path.expanduser("~")
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(ROOT_DIR, "data")
OUTPUT_PATH = os.path.join(USER_DIR, ".nuaal", "outputs")
LOG_PATH = os.path.join(USER_DIR, ".nuaal", "logs")
TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S"



from nuaal.utils import get_logger
from nuaal.connections.cli import Cisco_IOS_Cli
import queue
import threading
import time

class CliMultiRunner(object):
    """
    This class allows running set of CLI commands on multiple devices in parallel, using Worker threads
    """
    def __init__(self, provider, ips, actions=None, workers=4, DEBUG=False):
        """

        :param dict provider: Dictionary with necessary info for creating connection
        :param list ips: List of IP addresses of the device
        :param list actions: List of actions to be run in each connection
        :param int workers: Number of worker threads to spawn
        :param bool DEBUG: Enables/disables debugging output
        """
        self.provider = provider
        self.ips = ips
        self.workers = workers
        self.DEBUG = DEBUG
        self.logger = get_logger(name="CliMultiRunner", DEBUG=self.DEBUG)
        self.queue = queue.Queue()
        self.threads = []
        self.actions = actions if isinstance(actions, list) else []
        self.data = []

    def fill_queue(self):
        """
        This function builds *device-specific* providers for CliConnection and adds them to Queue object.

        :return: ``None``
        """
        for i in range(len(self.ips)):
            provider = dict(self.provider)
            provider["ip"] = self.ips[i]
            self.queue.put(provider)

    def worker(self):
        """
        Worker function to handle individual connections. Based on provider object from Queue establishes connection to device and runs
        defined set of commands. Received data are stored in ``self.data`` as a list of dictionaries, each dictionary representing one device.

        :return: ``None``
        """
        self.logger.debug(msg="Spawned new worker in {}".format(threading.current_thread()))
        while not self.queue.empty():
            provider = self.queue.get()
            if provider is None:
                break
            try:
                with Cisco_IOS_Cli(**provider) as device:
                    if "get_vlans" in self.actions:
                        device.get_vlans()
                    if "get_neighbors" in self.actions:
                        device.get_neighbors()
                    if "get_interfaces" in self.actions:
                        device.get_interfaces()
                    if "get_trunks" in self.actions:
                        device.get_trunks()
                    if "get_portchannels" in self.actions:
                        device.get_portchannels()
                    if "get_version" in self.actions:
                        device.get_version()
                    if "get_license" in self.actions:
                        device.get_license()
                    if "get_inventory" in self.actions:
                        device.get_inventory()
                    if "get_config" in self.actions:
                        device.get_config()
                    self.data.append(device.data)
            except Exception as e:
                self.logger.error(msg="Unhandled Exception occurred in thread '{}'. Exception: {}".format(threading.current_thread().getName(), repr(e)))
            self.queue.task_done()

    def thread_factory(self):
        """
        Function for spawning worker threads based on number of workers in ``self.workers``

        :return:
        """
        for i in range(self.workers):
            t = threading.Thread(name="WorkerThread-{}".format(i), target=self.worker, args=())
            self.threads.append(t)

    def run(self):
        """
        Main entry function, puts all pieces together. Firs populates Queue with IP addreses and connection information, spawns threads and starts them.
        Blocks until ``self.queue`` is empty.

        :return:
        """
        self.fill_queue()
        self.thread_factory()
        [t.start() for t in self.threads]
        self.queue.join()



from nuaal.Models.BaseModels import BaseModel, DeviceBaseModel
from nuaal.connections.api.apic_em.ApicEmBase import ApicEmBase
from nuaal.utils import Filter
import copy

class ApicEmDeviceModel(DeviceBaseModel):
    """

    """
    def __init__(self, apic=None, object_id=None, filter=None, DEBUG=False):
        """

        :param apic:
        :param object_id:
        :param filter:
        :param DEBUG:
        """
        super(ApicEmDeviceModel, self).__init__(name="ApicEmDeviceModel", DEBUG=DEBUG)
        self.apic = apic if isinstance(apic, ApicEmBase) else ApicEmBase()
        self.filter = filter
        self.apic._initialize()
        self.apic_object_id = object_id
        self._initialize()

    def _initialize(self):
        """

        :return:
        """
        if self.apic_object_id is None:
            if "id" in self.filter.required.keys():
                self.apic_object_id = self.filter.required["id"]
            else:
                self.logger.debug(msg="No apic_object_id provided, trying to match based on filter.")
                try:
                    response = self.apic.get(path="/network-device")
                    response = self.filter.universal_cleanup(data=response)
                    if len(response) == 1:
                        self.logger.debug(msg="Exactly one object matched query. apic_object_id: '{}'".format(response[0]["id"]))
                        self.apic_object_id = response[0]["id"]
                    else:
                        self.logger.error(msg="Multiple ({}) APIC-EM objects match filter query. Please provide more specific query or enter object_id manually.".format(len(response)))
                except Exception as e:
                    self.logger.critical(msg="Unhandled Exception occurred while trying to initialize. Exception: {}".format(repr(e)))
        response = self.apic.get(path="/network-device/{}".format(self.apic_object_id))
        print(response)
        self.device_info["mgmtIpAddress"] = response["managementIpAddress"]
        self.device_info["hostname"] = response["hostname"]
        self.device_info["vendor"] = "Cisco"
        self.device_info["platform"] = response["platformId"]
        self.device_info["swVersion"] = response["softwareVersion"]
        self.device_info["uptime"] = response["upTime"]

    def get_interfaces(self):
        """

        :return:
        """
        if self.apic_object_id is None:
            self.logger.error(msg="Cannot query APIC-EM for interfaces, no device ID found.")
            return {}
        response = self.apic.get(path="/interface/network-device/{}".format(self.apic_object_id))
        for interface in response:
            print(interface)
            name = interface["portName"]
            self.interfaces[name] = copy.deepcopy(self.interface_model)
            self.interfaces[name]["description"] = interface["description"],
            self.interfaces[name]["interfaceType"] = interface["interfaceType"],
            self.interfaces[name]["className"] = interface["className"],
            self.interfaces[name]["status"] = interface["status"],
            self.interfaces[name]["macAddress"] = interface["macAddress"].upper(),
            self.interfaces[name]["adminStatus"] = interface["adminStatus"],
            self.interfaces[name]["speed"] = interface["speed"],
            self.interfaces[name]["portName"] = interface["portName"],
            self.interfaces[name]["untaggedVlanId"] = interface["nativeVlanId"],
            self.interfaces[name]["taggedVlanIds"] = interface["vlanId"],
            self.interfaces[name]["duplex"] = interface["duplex"],
            self.interfaces[name]["portMode"] = interface["portMode"],
            self.interfaces[name]["portType"] = interface["portType"],
            self.interfaces[name]["ipv4Mask"] = interface["ipv4Mask"],
            self.interfaces[name]["ipv4Address"] = interface["ipv4Address"],
            self.interfaces[name]["mediaType"] = interface["mediaType"],

        return self.interfaces

    def get_vlans(self):
        """

        :return:
        """
        raise NotImplemented
        if self.apic_object_id is None:
            self.logger.error(msg="Cannot query APIC-EM for interfaces, no device ID found.")
            return {}
        response = self.apic.get(path="/network-device/{}/vlan".format(self.apic_object_id))
        for vlan in response:
            print(response)

    def get_inventory(self):
        """

        :return:
        """
        if self.apic_object_id is None:
            self.logger.error(msg="Cannot query APIC-EM for interfaces, no device ID found.")
            return {}
        response = self.apic.get(path="/network-device/module", params={"deviceId": self.apic_object_id})
        for raw_module in response:
            print(raw_module)
            module = copy.deepcopy(self.inventory_model)
            module["name"] = raw_module["name"]
            module["description"] = raw_module["description"]
            module["partNumber"] = raw_module["partNumber"]
            module["serialNumber"] = raw_module["serialNumber"]
            module["version"] = raw_module["assemblyRevision"]

            self.inventory.append(module)
        return self.inventory




from nuaal.connections.api.apic_em import ApicEmBase
from nuaal.utils import get_logger


class ApicEmPnpClient:
    def __init__(self, apic=None, DEBUG=False):
        """

        :param apic:
        :param DEBUG:
        """
        self.apic = apic
        self.DEBUG = DEBUG
        self.logger = get_logger(name="ApicEmPnpClient", DEBUG=self.DEBUG)
        if not isinstance(self.apic, ApicEmBase):
            self.logger.critical(msg="Given 'apic' is not an instance of 'ApicEmBase' class.")

    def template_get(self, id=None, filename=None):
        pass




from nuaal.connections.api.ciscoapi import CiscoApiBase
from nuaal.utils import get_logger
import json
from urllib.parse import quote
import timeit
from nuaal.definitions import DATA_PATH
from copy import deepcopy

class CiscoEoxApi(CiscoApiBase):
    def __init__(self, usernaeme=None, password=None, api_user_id=None, DEBUG=False):
        super(CiscoEoxApi, self).__init__(username=usernaeme,
                                          password=password,
                                          api_user_id=api_user_id,
                                          api_base_path="/supporttools/eox/rest/5",
                                          con_type="CiscoEoxApi",
                                          DEBUG=DEBUG
                                          )

    def process_response(self, response):
        eox_records = {}
        for eox_record in response:
            pass
        pass

    def eox_by_pn(self, pn):
        start = timeit.default_timer()
        data = []
        if isinstance(pn, str):
            pn = [x.strip() for x in pn.split(",")]
        self.logger.debug(msg="Getting EoX information for {} PNs".format(len(pn)))
        if len(pn) <= 20:
            pn_str = ",".join(pn)
            response = self.get(path="/{}/{}".format("EOXByProductID", pn_str))
            data += response["EOXRecord"]
        else:
            response = self.get_threaded(path="/EOXByProductID", queries=pn, workers=2, max_query_len=20)
            for x in response:
                data += x["EOXRecord"]
            print(len(data))
        end = timeit.default_timer()
        self.logger.debug(msg="Fetching {} PNs took {} seconds.".format(len(pn), end-start))
        return data

    def eox_by_sn(self, sn):
        eox_records = []
        if isinstance(sn, str):
            sn = [x.strip() for x in sn.split(",")]
        chunks = [sn[x:x + 20] for x in range(0, len(sn), 20)]
        for chunk in chunks:
            chunk = ",".join(chunk)
            response = self.get(path="/{}/{}".format("EOXBySerialNumber", chunk))
            eox_records += response["EOXRecord"]
        return eox_records

    def eox_by_date(self, start_date, end_date):
        response = self.get(path="/{}/{}/{}".format("EOXByDates", start_date, end_date))
        return response["EOXRecord"]

    def check_record(self, record):
        result = {"has_eox": None, "migration_pid": None, "is_wildcard": None}
        if record["EOLProductID"] == "":
            result["has_eox"] = False
            result["migration_pid"] = False
        else:
            result["has_eox"] = record["EOLProductID"]
            if record["EOXMigrationDetails"]["MigrationProductId"] == "":
                result["migration_pid"] = False
            else:
                result["migration_pid"] = record["EOXMigrationDetails"]["MigrationProductId"]
        if "*" in record["EOXInputValue"]:
            result["is_wildcard"] = True
        else:
            result["is_wildcard"] = False

        return result

    def process_response(self, eox_records, recursive=False):
        date_fields = ["EOXExternalAnnouncementDate", "EndOfSaleDate", "EndOfSWMaintenanceReleases", "EndOfRoutineFailureAnalysisDate",
                       "EndOfServiceContractRenewal", "LastDateOfSupport", "EndOfSvcAttachDate", "UpdatedTimeStamp"]
        structure = {
            "EOLProductID": "",
            "ProductIDDescription": "",
            "ProductBulletinNumber": "",
            "LinkToProductBulletinURL": "",
            "EOXExternalAnnouncementDate": "",
            "EndOfSaleDate": "",
            "EndOfSWMaintenanceReleases": "",
            "EndOfRoutineFailureAnalysisDate": "",
            "EndOfServiceContractRenewal": "",
            "LastDateOfSupport": "",
            "EndOfSvcAttachDate": "",
            "UpdatedTimeStamp": "",
            "EOXMigrationDetails": {}
        }
        migration_details = {
            "PIDActiveFlag": "",
            "MigrationInformation": "",
            "MigrationOption": "",
            "MigrationProductId": "",
            "MigrationProductName": "",
            "MigrationStrategy": "",
            "MigrationProductInfoURL": "",
            "PIDStatus": "Unknown"
            }

        new_records = {}
        for eox_record in eox_records:
            result = self.check_record(eox_record)
            if result["has_eox"]:
                self.logger.debug(msg="Found EoX info for PN: {}".format(result["has_eox"]))
                pid = result["has_eox"]
                if pid not in new_records.keys():
                    new_record = deepcopy(structure)
                    for key in structure.keys():
                        if isinstance(eox_record[key], str):
                            new_record[key] = eox_record[key]
                        elif key in date_fields:
                            new_record[key] = eox_record[key]["value"]
                        else:
                            pass
                    new_records[pid] = new_record
                if result["migration_pid"]:
                    new_migration = deepcopy(migration_details)
                    for key, value in eox_record["EOXMigrationDetails"].items():
                        new_migration[key] = value
                    if result["migration_pid"] not in new_records[pid]["EOXMigrationDetails"].keys():
                        new_records[pid]["EOXMigrationDetails"][result["migration_pid"]] = new_migration
                print(new_record)
            else:
                self.logger.debug(msg="No EoX records found for query: {} : {}".format(eox_record["EOXInputType"], eox_record["EOXInputValue"]))
                if not result["is_wildcard"]:
                    pass
                    # TODO: Continue here
        return new_records

    def recursive_check(self, eox_records):
        pids_to_check = []
        while True:
            eox_pids = list(eox_records.keys())
            for pid, record in eox_records.items():
                for migration_pid, migration_record in record["EOXMigrationDetails"].items():
                    if migration_record["PIDStatus"] == "Unknown":
                        if migration_pid not in eox_records.keys():
                            pids_to_check.append(migration_pid)
                        else:
                            migration_record["PIDStatus"] = "FoundEoX"
            print(pids_to_check)
            if len(pids_to_check) == 0:
                break
            eox_records.update(self.process_response(self.eox_by_pn(pn=pids_to_check)))
            pids_to_check = []







if __name__ == '__main__':
    eoxapi = CiscoEoxApi(api_user_id="ad", DEBUG=True)
    eoxapi._initialize()
    short = ["WS-C3850-24T-L"]
    long = ["WS-C3560G-48PS-S", "WS-C3850-24T-L","WS-C3850-48T-L","WS-C3850-24P-L","WS-C3850-24U-L","WS-C3850-48P-L","WS-C3850-48F-L","WS-C3850-48U-L","WS-C3850-24T-S","WS-C3850-48T-S","WS-C3850-24P-S","WS-C3850-24U-S","WS-C3850-48P-S","WS-C3850-48F-S","WS-C3850-48U-S","WS-C3850-24T-E","WS-C3850-48T-E","WS-C3850-24P-E","WS-C3850-24U-E","WS-C3850-48P-E","WS-C3850-48F-E","WS-C3850-48U-E","WS-C3850-12X48U-L","WS-C3850-12X48U-S","WS-C3850-12X48U-E","WS-C3850-24XU-L","WS-C3850-24XU-S","WS-C3850-24XU-E","WS-C3850-12S-S","WS-C3850-12S-E","WS-C3850-24S-S","WS-C3850-24S-E","WS-C3850-12XS-S","WS-C3850-12XS-E","WS-C3850-24XS-S","WS-C3850-24XS-E","WS-C3850-48XS-S","WS-C3850-48XS-E","WS-C3850-48XS-F-S","WS-C3850-48XS-F-E"]
    #with open("{}/{}".format(DATA_PATH, "eox_test.json"), mode="w") as f:
        #json.dump(obj=eoxapi.eox_by_pn(long), fp=f, indent=2)
    #print(json.dumps(eoxapi.eox_by_pn(long), indent=2))
    eox_records = None
    with open("{}/{}".format(DATA_PATH, "eox_test.json"), mode="r") as f:
        eox_records = json.load(f)
    eox_records = eoxapi.process_response(eox_records=eox_records)
    eoxapi.recursive_check(eox_records)


from nuaal.connections.cli import CliMultiRunner
from nuaal.discovery.Topology import CliTopology
from nuaal.utils import get_logger
from nuaal.Parsers import CiscoIOSParser
import json

class IP_Discovery:
    """
    This function performs discovery of the network devices based on their IP addresses.
    """
    def __init__(self, provider, DEBUG=False):
        """

        :param dict provider: Provider dictionary containing information for creating connection object, such as credentials
        :param bool DEBUG: Enables debugging output
        """
        self.DEBUG = DEBUG
        self.logger = get_logger(name="IP_Discovery", DEBUG=self.DEBUG)
        self.provider = provider
        self.data = None
        self.topology = None

    def run(self, ips):
        """
        Main entry function. Starts the discovery process based on given IP address of the seed device.

        :param list ips: List of IP addresses for discovery
        :return: ``None``
        """
        runner = CliMultiRunner(provider=self.provider, ips=ips, workers=6)
        runner.actions = ["get_neighbors"]
        runner.run()
        self.data = runner.data
        topo = CliTopology()
        topo.build_topology(self.data)
        self.topology = topo.topology



from nuaal.connections.api.ciscoapi import CiscoApiBase
import json
from copy import deepcopy


class EoxApi(CiscoApiBase):
    def __init__(self, usernaeme=None, password=None, api_user_id=None, DEBUG=False):
        super(EoxApi, self).__init__(username=usernaeme,
                                     password=password,
                                     api_user_id=api_user_id,
                                     api_base_path="/supporttools/eox/rest/5",
                                     con_type="CiscoEoxApi",
                                     DEBUG=DEBUG
                                     )
        self.data = {}
        self._initialize()

    def clear_query(self, query):
        if isinstance(query, str):
            query = [x.strip() for x in query.split(",")]
        elif isinstance(query, list):
            query = [x.strip() for x in query]
        else:
            raise TypeError("clear_query takes CSV-like string or list of strings "
                            "as argument. {} given.".format(type(query)))
        return query

    def get_by_query(self, path, query, max_queries=20, workers=2):
        raw_response = []
        try:
            query = self.clear_query(query)
        except TypeError:
            self.logger.error("Unsupported query format. Use either CSV-like string or list of srings.")
            query = []
        if len(query) <= max_queries:
            query_str = ",".join(query)
            try:
                raw_response = self.get(path="/{}/{}".format(path, query_str))["EOXRecord"]
            except Exception as e:
                self.logger.error(msg="Encountered Exception when retrieving EoX Response: {}".format(repr(e)))
                raw_response = []
            print(json.dumps(raw_response, indent=2))
        else:
            raw_response = self.get_threaded(path=path, queries=query, max_query_len=max_queries, workers=workers)
            new_response = []
            try:
                raw_response = [x["EOXRecord"] for x in raw_response]
                for page in raw_response:
                    new_response += page
                raw_response = new_response
            except Exception as e:
                self.logger.error(msg="Encountered Exception when retrieving EoX Response: {}".format(repr(e)))
                raw_response = []
            print(json.dumps(raw_response, indent=2))

        return self.process_response(query=query, raw_response=raw_response)

    def process_response(self, query, raw_response):
        date_fields = ["EOXExternalAnnouncementDate", "EndOfSaleDate", "EndOfSWMaintenanceReleases", "EndOfRoutineFailureAnalysisDate",
                       "EndOfServiceContractRenewal", "LastDateOfSupport", "EndOfSvcAttachDate", "UpdatedTimeStamp"]
        structure = {
            "EOLProductID": "",
            "ProductIDDescription": "",
            "ProductBulletinNumber": "",
            "LinkToProductBulletinURL": "",
            "EOXExternalAnnouncementDate": "",
            "EndOfSaleDate": "",
            "EndOfSWMaintenanceReleases": "",
            "EndOfRoutineFailureAnalysisDate": "",
            "EndOfServiceContractRenewal": "",
            "LastDateOfSupport": "",
            "EndOfSvcAttachDate": "",
            "UpdatedTimeStamp": "",
            "EOXMigrationDetails": {},
            "EOXInputType": "",
            "EOXInputValue": ""
        }
        migration_details = {
            "PIDActiveFlag": "",
            "MigrationInformation": "",
            "MigrationOption": "",
            "MigrationProductId": "",
            "MigrationProductName": "",
            "MigrationStrategy": "",
            "MigrationProductInfoURL": "",
            "PIDStatus": "Unknown"
        }
        processed_records = {}
        for eox_record in raw_response:
            if eox_record["EOLProductID"] == "":
                continue
            if eox_record["EOLProductID"] not in processed_records.keys():
                new_record = deepcopy(structure)
                for key in structure:
                    if isinstance(eox_record[key], str):
                        new_record[key] = eox_record[key]
                    elif key in date_fields:
                        new_record[key] = eox_record[key]["value"]
                    else:
                        pass
                processed_records[eox_record["EOLProductID"]] = new_record
            if eox_record["EOXMigrationDetails"]["MigrationProductId"] != "":
                new_migration = deepcopy(migration_details)
                for k, v in eox_record["EOXMigrationDetails"].items():
                    new_migration[k] = v
                new_migration["PIDStatus"] = "Unknown"
                processed_records[eox_record["EOLProductID"]]["EOXMigrationDetails"][eox_record["EOXMigrationDetails"]["MigrationProductId"]] = new_migration
        return processed_records






if __name__ == '__main__':
    long = ["WS-C3560G-48PS-S", "WS-C3850-24T-L","WS-C3850-48T-L","WS-C3850-24P-L","WS-C3850-24U-L","WS-C3850-48P-L","WS-C3850-48F-L","WS-C3850-48U-L","WS-C3850-24T-S","WS-C3850-48T-S","WS-C3850-24P-S","WS-C3850-24U-S","WS-C3850-48P-S","WS-C3850-48F-S","WS-C3850-48U-S","WS-C3850-24T-E","WS-C3850-48T-E","WS-C3850-24P-E","WS-C3850-24U-E","WS-C3850-48P-E","WS-C3850-48F-E","WS-C3850-48U-E","WS-C3850-12X48U-L","WS-C3850-12X48U-S","WS-C3850-12X48U-E","WS-C3850-24XU-L","WS-C3850-24XU-S","WS-C3850-24XU-E","WS-C3850-12S-S","WS-C3850-12S-E","WS-C3850-24S-S","WS-C3850-24S-E","WS-C3850-12XS-S","WS-C3850-12XS-E","WS-C3850-24XS-S","WS-C3850-24XS-E","WS-C3850-48XS-S","WS-C3850-48XS-E","WS-C3850-48XS-F-S","WS-C3850-48XS-F-E"]
    short = ["WS-C3560G-48PS-S"]
    eoxapi = EoxApi(api_user_id="an", DEBUG=True)
    print(json.dumps(eoxapi.get_by_query(path="EOXByProductID", query=long), indent=2))



from nuaal.connections.cli import CliBaseConnection
from nuaal.Parsers import CiscoIOSParser
from nuaal.utils import vlan_range_expander, vlan_range_shortener
from nuaal.definitions import DATA_PATH
import json
import threading
import queue
import datetime

class Cisco_IOS_Cli(CliBaseConnection):
    """
    Object for interaction with network devices running Cisco IOS (or IOS XE) software via CLI interface.
    """
    def __init__(self, ip=None, username=None, password=None, parser=None, secret=None, method="ssh", enable=False, store_outputs=False, DEBUG=False):
        """

        :param ip: (str) IP address or FQDN of the device you're trying to connect to
        :param username: (str) Username used for login to device
        :param password: (str) Password used for login to device
        :param parser: (ParserModule) Instance of ParserModule class which will be used for parsing of text outputs. By default, new instance of ParserModule is created.
        :param secret: (str) Enable secret for accessing Privileged EXEC Mode
        :param method: (str) Primary method of connection, 'ssh' or 'telnet'. (Default is 'ssh')
        :param enable: (bool) Whether or not enable Privileged EXEC Mode on device
        :param store_outputs: (bool) Whether or not store text outputs of sent commands
        :param DEBUG: (bool) Enable debugging logging
        """
        super(Cisco_IOS_Cli, self).__init__(ip=ip, username=username, password=password,
                                            parser=parser if isinstance(parser, CiscoIOSParser) else CiscoIOSParser(DEBUG=False),
                                            secret=secret, enable=enable, store_outputs=store_outputs, DEBUG=DEBUG)
        self.prompt_end = [">", "#"]
        self.ssh_method = "cisco_ios"
        self.telnet_method = "cisco_ios_telnet"
        self.primary_method = self.ssh_method if method == "ssh" else self.telnet_method
        self.secondary_method = self.telnet_method if method == "ssh" else self.ssh_method
        self.command_mappings = {
            "get_vlans": [
                "show vlan brief",
                "show vlan-switch brief"
            ],
            "get_mac_address_table": [
                "show mac address-table",
                "show mac-address-table"
            ],
            "get_neighbors": [
                "show cdp neighbors detail"
            ],
            "get_inventory": [
                "show inventory"
            ],
            "get_interfaces": [
                "show interfaces"
            ],
            "get_portchannels": [
                "show etherchannel summary"
            ],
            "get_arp": [
                "show ip arp"
            ],
            "get_license": [
                "show license"
            ],
            "get_version": [
                "show version"
            ]
        }

    def get_neighbors(self, output_filter=None, strip_domain=False):
        """
        Function to get neighbors of the device with possibility to filter neighbors
        :param output_filter: (Filter) Instance of Filter class, used to filter neighbors, such as only "Switch" or "Router"
        :param strip_domain: (bool) Whether or not to strip domain names and leave only device hostname
        :return: List of dictionaries
        """
        command = "show cdp neighbors detail"
        raw_output = self._send_command(command=command)
        if not raw_output:
            return []
        if self.store_outputs:
            self.store_raw_output(command=command, raw_output=raw_output)
        parsed_output = self.parser.autoparse(text=raw_output, command=command)
        if output_filter:
            parsed_output = output_filter.universal_cleanup(data=parsed_output)
        if strip_domain:
            for neighbor in parsed_output:
                neighbor["hostname"] = neighbor["hostname"].split(".")[0]
        self.data["neighbors"] = parsed_output
        return parsed_output

    def get_trunks(self, expand_vlan_groups=False):
        """
        Custom parsing function for output of "show interfaces trunk"
        :param expand_vlan_groups: (bool) Whether or not to expand VLAN ranges, for example '100-102' -> [100, 101, 102]
        :return: List of dictionaries
        """
        command = "show interfaces trunk"
        raw_output = self._send_command(command=command)
        if len(raw_output) == 0:
            self.data["trunk_interfaces"] = []
            return []
        if self.store_outputs:
            self.store_raw_output(command=command, raw_output=raw_output)
        parsed_output = self.parser.trunk_parser(text=raw_output)
        if expand_vlan_groups:
            for trunk in parsed_output:
                trunk["allowed"] = vlan_range_expander(trunk["allowed"])
                trunk["active"] = vlan_range_expander(trunk["active"])
                trunk["forwarding"] = vlan_range_expander(trunk["forwarding"])
        self.data["trunk_interfaces"] = parsed_output
        return parsed_output

    def get_config(self):
        """
        Function for retrieving current configuration of the device.

        :return str: Device configuration
        """
        command = "show running-config"
        raw_output = self._send_command(command=command)
        if self.store_outputs:
            self.store_raw_output(command=command, raw_output=raw_output)
        return raw_output



from nuaal.Writers import Writer
from nuaal.utils import check_path
from nuaal.definitions import DATA_PATH
import xlsxwriter

class ExcelWriter(Writer):
    """
    This class provides functions for writing the retrieved data in Microsoft Excel format, `.xlsx`.
    """
    def __init__(self):
        super(ExcelWriter, self).__init__(type="Excel")

    def create_workbook(self, path, filename):
        """
        Function for creating Excel Workbook based on path and file name.

        :param str path: System path to Excel file
        :param str filename: Name of the Excel file.
        :return: Instance of ``xlsxwriter`` workbook object.
        """
        path = check_path(path)
        workbook = xlsxwriter.Workbook("{}/{}".format(path, filename))
        return workbook

    def write_list(self, workbook, data, worksheetname=None, headers=None):
        """
        Function for creating worksheet inside given workbook and writing provided data.

        :param workbook: Reference of the ``xlsxwriter`` workbook object.
        :param list data: List of lists, where each list represents one row in the worksheet.
        :param str worksheetname: Name of the worksheet.
        :param list headers: List of column headers.
        :return: ``None``
        """
        if isinstance(data, list):
            if worksheetname:
                self.logger.info(msg="Creating new worksheet with name {}.".format(worksheetname))
                worksheet = workbook.add_worksheet(name=worksheetname)
            else:
                self.logger.info(msg="Creating new worksheet.")
                worksheet = workbook.add_worksheet()
            row_pointer = 0
            column_pointer = 0
            bold = workbook.add_format({"bold": True})
            if headers:
                worksheet.write_row(row_pointer, column_pointer, headers, bold)
                row_pointer += 1
            for entry in data:
                if isinstance(entry, list):
                    worksheet.write_row(row_pointer, column_pointer, entry)
                    row_pointer += 1
            if headers:
                worksheet.autofilter(0, 0, row_pointer-1, len(data[0])-1)
            self.logger.info(msg="{} entries written.".format(len(data)))

    def write_json(self, workbook, data, worksheetname=None, headers=None):
        """
        Function for writing JSON-like data to worksheet.

        :param workbook: Reference of the ``xlsxwriter`` workbook object.
        :param data:
        :param str worksheetname: Name of the worksheet
        :param list headers: List of column headers.
        :return: ``None``
        """
        if isinstance(data, list):
            if not headers:
                # Generate headers from first dict
                headers = list(data[0].keys())
            bold = workbook.add_format({"bold": True})
            # Create new worksheet
            worksheet = None
            if worksheetname:
                self.logger.info(msg="Creating new worksheet with name {}.".format(worksheetname))
                worksheet = workbook.add_worksheet(name=worksheetname)
            else:
                self.logger.info(msg="Creating new worksheet.")
                worksheet = workbook.add_worksheet()
            row_pointer = 0
            column_pointer = 0
            worksheet.write_row(row_pointer, column_pointer, headers, bold)
            row_pointer += 1
            for entry in data:
                row = []
                # Workaround for unsupported types
                for element in [entry[x] for x in headers]:
                    if isinstance(element, list):
                        row.append(str(element))
                    else:
                        row.append(element)
                worksheet.write_row(row_pointer, column_pointer, row)
                row_pointer += 1

            self.logger.info(msg="{} entries writen.".format(len(data)))

        else:
            # TODO: Implement other JSON types
            self.logger.error(msg="Given JSON is not a list of dictionaries. Not yet implemented.")

    def write_data(self, workbook, data):
        """
        Function for writing entire content of device data, divided into sections based on `get_` command used.
        :param workbook: Reference of the ``xlsxwriter`` workbook object.
        :param dict data: Content of device.data variable, which holds all retrieved info.
        :return: ``None``
        """
        headers, content = self.combine_device_data(data)
        for section in headers:
            self.write_list(workbook=workbook, data=content[section], worksheetname=section, headers=headers[section])



from nuaal.utils import get_logger
from nuaal.utils import Filter
from nuaal.connections.cli import Cisco_IOS_Cli
from nuaal.discovery.Topology import CliTopology
import threading
from nuaal.definitions import DATA_PATH
import json
import queue
import timeit
from datetime import datetime

class Neighbor_Discovery:
    """
    This class provides a simple way to perform network discovery based on CDP neighbors of device.
    Given IP address of initial device (or 'seed device') it tries to crawl trough ne network and discover all supported devices.
    CDP must be enabled on devices.
    """
    def __init__(self, provider, DEBUG=False):
        """

        :param dict provider: Provider dictionary containing information for creating connection object, such as credentials
        :param bool DEBUG: Enables/disables debugging output
        """
        self.DEBUG = DEBUG
        self.logger = get_logger(name="NeighborDiscovery", DEBUG=self.DEBUG)
        self.provider = provider
        self.data = {}
        self.topology = None
        self.visited = []
        self.discovered = []
        self.to_process = []
        self.to_visit = []
        self.workers = 4
        self.queue = queue.Queue()
        self.threads = []
        self.current_id = 0
        self.neighbor_filter = Filter(required={"capabilities": ["Router", "Switch"]}, exact_match=False)

    def _gen_device_id(self, ip, hostname=None):
        """
        Function to generate pseudo-unique identification of device. By default, it returns device hostname,
        if hostname is one of the default hostnames, it uses IP address to distinguish devices with same hostname.

        :param ip: String - IP address of the device
        :param hostname: String (optional) - Hostname of the device
        :return: String - such as "C2960X_AB01" or "Router(192.168.1.1)"
        """
        if not hostname:
            return "({})".format(ip)
        default_hostnames = ["Router", "Switch"]
        if hostname not in default_hostnames:
            return hostname
        else:
            return "{}({})".format(hostname, ip)

    def get_neighbors(self, ip, hostname=None):
        """
        This function performs the discovery of the neighbors for given device. Results are stored in self.to_process.
        After each round, the results are processed by self.process_neigbors()

        :param ip: String - IP Address of the device
        :param hostname: String (optinal) - Hostname of the device, if not given, will be set after connecting to device
        :return: ``None``
        """
        device_id = self._gen_device_id(ip=ip)
        start_time = timeit.default_timer()
        if hostname:
            device_id = self._gen_device_id(ip=ip, hostname=hostname)
        device_neighbors = []
        self.logger.info(msg="Discovering device {}".format(device_id))
        with Cisco_IOS_Cli(ip=ip, **self.provider) as device:
            if not device.device:
                self.logger.error(msg="Could not connect to device {}. Failed after {} seconds.".format(device_id, timeit.default_timer() - start_time))
                self.to_process.append({device_id: []})
            else:
                device_id = self._gen_device_id(ip=ip, hostname=device.data["hostname"])
                if device_id not in self.discovered:
                    self.logger.warning(msg="Device {} is being visited, but it is not in discovered. This happens only for seed device.".format(device_id))
                    self.discovered.append(device_id)
                try:
                    device_neighbors = device.get_neighbors(output_filter=self.neighbor_filter, strip_domain=True)
                except Exception as e:
                    self.logger.error(msg="Could not retrieve neighbors of {}. Reason: {}. {} seconds.".format(device_id, repr(e), timeit.default_timer() - start_time))
                finally:
                    if {device_id: device_neighbors} in self.to_process:
                        print("THIS SHOULD NOT HAPPEN!")
                    self.logger.debug(msg="Storing results from {} for later processing. Time: {} seconds.".format(device_id, timeit.default_timer() - start_time))
                    self.to_process.append({device_id: device_neighbors})
                    try:
                        self.to_visit.remove({device_id: {"ip": ip, "hostname": hostname}})
                    except ValueError:
                        self.logger.debug(msg="Could not remove device {} from self.to_visit. Item not in list.".format(device_id))
                    except Exception as e:
                        print(repr(e))

    def process_neighbors(self):
        """
        This functions decides what to do with neighbors of device, such as which were already discovered or visited.
        It runs after every discovery round to combine data retrieved by individual worker threads.

        :return: ``None``
        """
        start_time = timeit.default_timer()
        for item in self.to_process:
            # 'item' represents neighbors of device
            for device_id, neighbors in item.items():
                if device_id in self.visited:
                    print("This should not happen!")
                else:
                    self.visited.append(device_id)
                    self.logger.debug(msg="Device {} has been visited.".format(device_id))
                self.logger.info(msg="Processing {} neighbor(s) of device {}".format(len(neighbors), device_id))
                self.data[device_id] = neighbors
                for neighbor in neighbors:
                    neighbor_id = self._gen_device_id(ip= neighbor["ipAddress"], hostname=neighbor["hostname"])
                    if neighbor_id in self.visited:
                        self.logger.info(msg="Neighbor {} of device {} has already been visited.".format(neighbor_id, device_id))
                        continue
                    elif neighbor_id in self.discovered:
                        self.logger.info(msg="Neighbor {} of device {} has already been discovered, waiting to be visited.".format(neighbor_id, device_id))
                        continue
                    else:
                        self.logger.info(msg="Discovered new neighbor {} of device {}.".format(neighbor_id, device_id))
                        self.discovered.append(neighbor_id)
                        self.to_visit.append({neighbor_id: {"ip": neighbor["ipAddress"], "hostname": neighbor["hostname"]}})
                        self.queue.put({"ip": neighbor["ipAddress"], "hostname": neighbor["hostname"]})
        self.to_process = []
        self.logger.info(msg="All outputs of current round have been processed. Time: {} seconds.".format(timeit.default_timer() - start_time))

    def worker(self):
        """
        This is a wrapper function that is run as thread.

        :return: ``None``
        """
        self.logger.debug(msg="Thread {} started.".format(threading.current_thread().getName()))
        if self.queue.empty():
            self.logger.debug(msg="No work for thread {}.".format(threading.current_thread().getName()))
        while not self.queue.empty():
            params = self.queue.get()
            if params is None:
                break
            self.get_neighbors(**params)
            self.queue.task_done()

    def run(self, ip):
        """
        This function starts the discovery process based on given IP address of the `seed` device.

        :param str ip: IP address of the seed device
        :return: ``None``
        """
        self.get_neighbors(ip=ip)
        self.process_neighbors()
        while len(self.to_visit) > 0:
            try:
                for i in range(self.workers):
                    t = threading.Thread(name="DiscoveryThread-{}".format(i), target=self.worker)
                    self.threads.append(t)
                [t.start() for t in self.threads]
                self.queue.join()
                self.threads = []
                self.process_neighbors()
            except KeyboardInterrupt:
                self.logger.info(msg="Keyboard Interrupt")
                break
            finally:
                self.process_neighbors()
        try:
            with open("{}\discovery\{}-{}.json".format(DATA_PATH, ip, str(datetime.now()).replace(':', '-')), mode="w") as f:
                json.dump(self.data, f, indent=2)
        except FileNotFoundError:
            self.logger.error(msg="Could not write discovery results to file. Reason: File does not exist.")
        topo = CliTopology()
        topo.build_topology(self.data)
        self.topology = topo.topology





from nuaal.Parsers.PatternsLib import PatternsLib
from nuaal.Parsers.Parser import ParserModule
from nuaal.Parsers.CiscoIOSParser import CiscoIOSParser
from nuaal.Parsers.JuniperJUNOSParser import JuniperJUNOSParser


from nuaal.Parsers import ParserModule
import timeit
import re

class CiscoIOSParser(ParserModule):
    """
    Child class of `ParserModule` designed for `cisco_ios` device type.
    """
    def __init__(self, DEBUG=False):
        super(CiscoIOSParser, self).__init__(device_type="cisco_ios", DEBUG=DEBUG)

    def vlanGroup_check(self, vlanGroup):
        if isinstance(vlanGroup, list):
            if len(vlanGroup) > 0:
                vlanGroup = vlanGroup[0]
            else:
                vlanGroup = []
                return vlanGroup
        if isinstance(vlanGroup, str):
            if vlanGroup == "none":
                vlanGroup = []
                return vlanGroup
            vlanGroup = vlanGroup.split(",")
        elif isinstance(vlanGroup, int):
            vlanGroup = [str(vlanGroup)]
        return vlanGroup

    def trunk_parser(self, text):
        """
        Function specifically designed for parsing output of `show interfaces trunk` command.

        :param str text: Plaintext output of `show interfaces trunk` command.
        :return: List of dictionaries representing trunk interfaces.
        """
        start_time = timeit.default_timer()
        section_pattern = re.compile(pattern=r"(?:^Port.*$(?:\s^\S.*$)+)", flags=re.MULTILINE)
        sections = self.match_single_pattern(pattern=section_pattern, text=text)
        if len(sections) != 4:
            return []
        mode_pattern = re.compile(
            pattern=r"^(?P<interface>[A-Za-z]+\d+(\/\d+){0,2})\s+(?P<mode>\S+)\s+(?P<encapsulation>\S+)\s+(?P<status>\S+)\s+(?P<nativeVlan>\d+)",
            flags=re.MULTILINE
        )
        interface_pattern = re.compile(pattern=r"^(?P<interface>[A-Za-z]+\d+(\/\d+){0,2})\s+(?P<vlanGroup>(?:\d{1,4}(?:-\d{1,4})?,?)+)", flags=re.MULTILINE)
        mode = self.match_single_pattern(text=sections[0], pattern=mode_pattern)
        allowed = self.match_single_pattern(text=sections[1], pattern=interface_pattern)
        active = self.match_single_pattern(text=sections[2], pattern=interface_pattern)
        forwarding = self.match_single_pattern(text=sections[3], pattern=interface_pattern)
        trunks = []
        for trunk in mode:
            entry = dict(trunk)
            entry["allowed"] = self.vlanGroup_check([x["vlanGroup"] for x in allowed if x["interface"] == entry["interface"]])
            entry["active"] = self.vlanGroup_check([x["vlanGroup"] for x in active if x["interface"] == entry["interface"]])
            entry["forwarding"] = self.vlanGroup_check([x["vlanGroup"] for x in forwarding if x["interface"] == entry["interface"]])
            trunks.append(entry)
        self.logger.info(msg="Parsing of 'show interfaces trunk' took {} seconds.".format((timeit.default_timer()-start_time)))
        return trunks



from nuaal.connections.cli import Cisco_IOS_Cli


def GetCliHandler(device_type=None, ip=None, username=None, password=None, parser=None, secret=None, enable=False, store_outputs=False, DEBUG=False):
    """
    This function can be used for getting the correct connection object for specific device type.

    :param str device_type: String representation of device type, such as `cisco_ios`
    :param ip: (str) IP address or FQDN of the device you're trying to connect to
    :param username: (str) Username used for login to device
    :param password: (str) Password used for login to device
    :param parser: (ParserModule) Instance of ParserModule class which will be used for parsing of text outputs. By default, new instance of ParserModule is created.
    :param secret: (str) Enable secret for accessing Privileged EXEC Mode
    :param enable: (bool) Whether or not enable Privileged EXEC Mode on device
    :param store_outputs: (bool) Whether or not store text outputs of sent commands
    :param DEBUG: (bool) Enable debugging logging
    :return: Instance of connection object
    """
    if device_type == "cisco_ios":
        return Cisco_IOS_Cli(ip=ip, username=username, password=password, parser=parser, secret=secret, enable=enable, store_outputs=store_outputs, DEBUG=DEBUG)



import re


class Patterns:
    """
    This class holds all the necessary regex patterns, which are used by `ParserModule`. This class returns compiled regex patterns based on specified device
    type.
    """
    def __init__(self, device_type, DEBUG=False):
        """

        :param str device_type: String representation of device type, such as `cisco_ios`
        :param bool DEBUG: Enables/disables debugging output
        """
        raise DeprecationWarning("This class is deprecated. Please use nuaal.Parsers.PatternsLib instead.")
        self.device_type = device_type
        self.DEBUG = DEBUG
        self.map = {
            "cisco_ios": self.cisco_ios_patterns()
        }

    def get_patterns(self):
        """
        Function used for retrieving compiled regex patterns.
        :return: Dictionary of compiled regex patterns
        """
        return self.map[self.device_type]

    def cisco_ios_patterns(self):
        """
        This function holds regex patterns for `cisco_ios` device type.
        :return:
        """
        patterns = {
            "level0": {
                "show inventory": [
                    re.compile(
                        pattern=r"^NAME:\s\"(?P<name>[\w\s()/]+)\",\s+DESCR:\s\"(?P<desc>(?:[\w\s(),.\-_+/#:&]+))\"\s*PID:\s(?P<pid>\S+)\s*,\s+VID:\s(?P<vid>\S+)?\s*,\s+SN:\s+(?P<sn>\S+)",
                        flags=re.MULTILINE
                    )
                ],
                "show vlan brief": [
                    re.compile(
                        pattern=r"^(?P<id>\d+)\s+(?P<name>\S+)\s+(?P<status>\S+)\s+(?P<access_ports>(?:[A-Za-z]+\d+(?:\/\d+){0,2},?\s+)+)?",
                        flags=re.MULTILINE
                    )
                ],
                "show interfaces": [
                    re.compile(
                        pattern=r"^\S.*(?:$\s+^\s.*)+",
                        flags=re.MULTILINE
                    )
                ],
                "show etherchannel summary": [
                    re.compile(
                        pattern="^(?P<group>\d+)\s+(?P<portchannel>Po\d{1,3})\((?P<status>[DIHRUPsSfMuwd]{1,2})\)\s+(?P<protocol>\S+)\s+(?P<ports>(?:(?:\w+\d+(?:\/\d+)*)\(\S\)\s*)+)",
                        flags=re.MULTILINE
                    )
                ],
                "show cdp neighbors detail": [
                    re.compile(
                        pattern=r"(?<=-{25}\n).*?(?=-{25}|$)",
                        flags=re.DOTALL
                    )
                ],
                "show version": [
                    re.compile(
                        pattern=r"^Cisco.*Configuration\sregister\sis\s\S+",
                        flags=re.DOTALL
                    )
                ],
                "show mac address-table": [
                    re.compile(
                        pattern=r"^\s+(?P<vlan>\S+)\s+(?P<mac>(?:[\da-f]{4}\.?){3})\s+(?P<type>\S+)\s+(?P<ports>\S+)",
                        flags=re.MULTILINE
                    ),
                    re.compile(
                        pattern=r"^(?P<mac>(?:[\da-f]{4}\.?){3})\s+(?P<type>\S+)\s+(?P<vlan>\S+)\s+(?P<ports>\S+)",
                        flags=re.MULTILINE
                    )
                ],
                "show ip arp": [
                    re.compile(
                        pattern=r"^(?P<protocol>\S+)\s+(?P<ipAddress>((?:\d{1,3}.?){4}))\s+(?P<age>(?:\d+|-))\s+(?P<mac>(?:[\da-f]{4}\.?){3})\s+(?P<type>\S+)\s+(?P<interface>\S+)",
                        flags=re.MULTILINE
                    )
                ],
                "show license": [
                    re.compile(
                        pattern=r"^Index.*(?:(?:$\s+^\s.*)+)?",
                        flags=re.MULTILINE
                    )
                ]
            },
            "level1": {
                "show vlan brief": {
                    "access_ports": [
                        re.compile(
                            pattern=r"[A-Za-z]+\d+(?:\/\d+){0,2}"
                        )
                    ]
                },
                "show version": {
                    "version": [
                        re.compile(
                            pattern=r"^(?P<vendor>[Cc]isco)\s(?P<software>IOS(?:\sXE)?)\sSoftware,.*Version\s(?P<version>[\w\.\(\)\-]+)(?:,)?",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^(?P<vendor>[Cc]isco)\s(?P<platform>[\w-]+).*with\s(?P<systemMemory>\d+K/\d+K)\sbytes\sof\smemory.",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^(?P<vendor>[Cc]isco)\s(?P<platform>[\w-]+).*with\s(?P<systemMemory>\d+K)\sbytes\sof\smemory.",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^$\s^(?P<hostname>[\w\-_]+)\suptime\sis\s(?P<uptime>.*)$",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^(?P<hostname>[\w\-_]+)\suptime\sis\s(?P<uptime>.*)$",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^System\simage\sfile\sis\s\"(?P<imageFile>.*)\"",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"Experimental\sVersion\s(?P<experimental_version>\S+)",
                            flags=re.MULTILINE
                        )
                    ]
                },
                "show interfaces": {
                    "name": [
                        re.compile(
                            pattern=r"^(?P<name>\S+)\sis\s(?P<status>.*),\sline\sprotocol\sis\s(?P<lineProtocol>\S+)",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^(?P<name>\S+)",
                            flags=re.MULTILINE
                        )
                    ],
                    "address": [
                        re.compile(
                            pattern=r"^\s+Hardware\sis\s(?P<hardware>.*),\saddress\sis\s(?P<mac>\S+)\s\(bia\s(?P<bia>\S+)\)",
                            flags=re.MULTILINE
                        )
                    ],
                    "description": [
                        re.compile(
                            pattern=r"^\s+Description:\s(?P<description>.*)",
                            flags=re.MULTILINE
                        )
                    ],
                    "ipv4Address": [
                        re.compile(
                            pattern=r"^\s+Internet\saddress\sis\s(?P<ipv4Address>[\d\.]+)\/(?P<ipv4Mask>\d+)",
                            flags=re.MULTILINE
                        )
                    ],
                    "rates": [
                        re.compile(
                            pattern=r"^\s+(?P<loadInterval>\d+\s\S+)\sinput\srate\s(?P<inputRate>\d+).*,\s(?P<inputPacketsInterval>\d+).*$"
                                                 r"\s+.*output\srate\s(?P<outputRate>\d+).*,\s(?P<outputPacketsInterval>\d+)",
                            flags=re.MULTILINE
                        )
                    ],
                    "duplex": [
                        re.compile(
                            pattern=r"^\s+(?P<duplex>\S+)-duplex,\s(?P<speed>(?:\d+)?\S+)(?:,\s+link\stype\sis\s(?P<linkType>\S+))?,\smedia\stype\sis\s(?P<mediaType>.*)",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^\s+(?P<duplex>\S+)-duplex,\s(?P<sped>\S+)$",
                            flags=re.MULTILINE
                        )
                    ],
                    "mtu": [re.compile(pattern=r"^\s+MTU\s(?P<mtu>\d+).*BW\s(?P<bandwidth>\d+)\sKbit(?:/sec)?,\sDLY\s(?P<delay>\d+).*$"
                                               r"\s+reliability\s(?P<reliability>\S+),\stxload\s(?P<txLoad>\S+),\srxload\s(?P<rxLoad>\S+)",
                                       flags=re.MULTILINE),
                            ],
                    "pseudowire": [
                        re.compile(
                            pattern=r"^\s+Encapsulation\s(?P<encapsulation>\w+)",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^\s+RX\s+(?P<rxPackets>\d+)\spackets\s(?P<rxBytes>\d+)\sbytes\s(?P<rxDrops>\d+)\sdrops\s+TX\s+(?P<txPackets>\d+)\spackets\s(?P<txBytes>\d+)\sbytes\s(?P<txDrops>\d+)\sdrops",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^\s+Peer\sIP\s(?P<peerIP>[\d\.]+),\sVC\sID\s(?P<virtualCircuitID>\d+)",
                            flags=re.MULTILINE
                        ),
                        re.compile(
                            pattern=r"^\s+MTU\s(?P<mtu>\d+)\sbytes",
                            flags=re.MULTILINE
                        )
                    ],
                    "input_counters": [
                        re.compile(pattern=r"^\s+(?P<rxPackets>\d+)\spackets\sinput,\s(?P<rxBytes>\d+)\sbytes,\s(?P<noBuffer>\d+)\sno\sbuffer$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+Received\s(?P<rxBroadcasts>\d+).*\((?P<rxMulticasts>\d+).*$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<runts>\d+)\srunts,\s(?P<giants>\d+)\sgiants,\s(?P<throttles>\d+).*$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<inputErrors>\d+)\sinput\serrors,\s(?P<crc>\d+)\sCRC,\s(?P<frame>\d)\sframe,\s(?P<overrun>\d+)\soverrun,\s(?P<ignored>\d+).*$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"(?:\s+(?P<watchdog>\d+)\swatchdog,\s(?P<multicasts>\d+)\smulticast,\s(?P<pauseInput>\d+)\spause\sinput$\s+(?P<inputPacketsWithDribbleCondition>\d+)\sinput.*)?",
                                   flags=re.MULTILINE)
                    ],
                    "output_counters": [
                        re.compile(pattern=r"^\s+(?P<txPackets>\d+)\spackets\soutput,\s(?P<txBytes>\d+)\sbytes,\s(?P<underruns>\d+)\sunderruns$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<outputErrors>\d+)\soutput\serrors,\s(?:(?P<collision>\d+)\scollisions,\s)?(?P<interfaceResets>\d+)\sinterface\sresets$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<babbles>\d+)\sbabbles,\s(?P<lateCollision>\d+)\slate\scollision,\s(?P<deferred>\d+)\sdeferred$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<lostCarrier>\d+)\slost\scarrier,\s(?P<noCarrier>\d+)\sno\scarrier,\s(?P<pauseOutput>\d+)\sPAUSE\soutput$",
                                   flags=re.MULTILINE),
                        re.compile(pattern=r"\s+(?P<outputBufferFailures>\d+)\soutput\sbuffer\sfailures,\s(?P<outputBufferSwappedOut>\d+)\soutput buffers swapped out$",
                                   flags=re.MULTILINE)
                    ]
                },
                "show etherchannel summary": {
                    "ports": [
                        re.compile(
                            pattern=r"(?P<port>\w+\d+(?:\/\d+)*)\((?P<status>[A-Za-z]+)\)"
                        )
                    ]
                },
                "show cdp neighbors detail": {
                    "hostname": [
                        re.compile(pattern=r"^Device\sID:\s(?P<hostname>[\w\_\-\(\)]+)(?:.\S+)?", flags=re.MULTILINE),
                        re.compile(pattern=r"^Device\sID:\s(?P<hostname>\S+)", flags=re.MULTILINE)
                    ],
                    "ipAddress": [re.compile(pattern=r"IP\saddress:\s(?P<ipAddress>(?:\d{1,3}\.?){4})", flags=re.MULTILINE)],
                    "platform": [re.compile(pattern=r"^Platform:\s(?:(?:Cisco|cisco\s)?(?P<platform>(?:\S+\s?)+))", flags=re.MULTILINE)],
                    "capabilities": [re.compile(pattern=r"Capabilities:\s(?P<capabilities>(?:\S+\s)+)")],
                    "localInterface": [re.compile(pattern=r"^Interface:\s(?P<localInterface>[A-Za-z]+\d+(?:\/\d+)*)", flags=re.MULTILINE)],
                    "remoteInterface": [re.compile(pattern=r"Port\sID\s\(outgoing\sport\):\s(?P<remoteInterface>[A-Za-z]+\d+(?:\/\d+)*)")]
                },
                "show license": {
                    "index": [re.compile(pattern=r"^Index\s(?P<index>\d+)")],
                    "feature": [re.compile(pattern=r"Feature:\s(?P<feature>\S+)")],
                    "period_left": [re.compile(pattern=r"Period\sleft:\s(?P<period_left>.*)")],
                    "period_used": [re.compile(pattern=r"Period\sUsed:\s(?P<period_used>.*)")],
                    "license_type": [re.compile(pattern=r"License\sType:\s(?P<license_type>.*)")],
                    "license_state": [re.compile(pattern=r"License\sState:\s(?P<license_state>.*)")],
                    "license_count": [re.compile(pattern=r"License\sCount:\s(?P<license_count>.*)")],
                    "license_priority": [re.compile(pattern=r"License\sPriority:\s(?P<license_priority>.*)")],
                }
            }
        }
        return patterns



from nuaal.connections.api import RestBase
from requests.auth import HTTPBasicAuth
import json
from nuaal.definitions import DATA_PATH

class EpnmBase(RestBase):
    def __init__(self, url, username, password, api_base_path, verify_ssl, DEBUG=False):
        super(EpnmBase, self).__init__(url=url, username=username, password=password, api_base_path=api_base_path, verify_ssl=verify_ssl, DEBUG=DEBUG)

    def _authorize(self):
        self.common_headers = {"headers": self.headers, "auth": HTTPBasicAuth(self.username, self.password), "verify": self.verify_ssl}
        self._store_credentials()

    def _store_credentials(self):
        self.logger.info(msg="Storing credentials.")
        with open(file="{}/epnm_credentials.json".format(DATA_PATH), mode="w") as f:
            json.dump({"username": self.username, "password": self.password}, f)

    def _load_credentials(self):
        with open(file="{}/epnm_credentials.json".format(DATA_PATH), mode="r") as f:
            credentials = json.load(fp=f)
            for value in credentials.values():
                if value == "":
                    return False
            self.username = credentials["username"]
            self.password = credentials["password"]
            self.common_headers = {"headers": self.headers, "auth": HTTPBasicAuth(self.username, self.password), "verify": self.verify_ssl}
            return True




from nuaal.connections.api import RestBase
from nuaal.definitions import DATA_PATH, TIMESTAMP_FORMAT
import requests
import json
from urllib.parse import quote
from threading import Thread
from queue import Queue
from datetime import datetime


class CiscoApiBase(RestBase):

    def __init__(self, url="https://api.cisco.com", username=None, password=None, api_user_id="default", api_base_path="", verify_ssl=False, con_type=None, DEBUG=False):
        super(CiscoApiBase, self).__init__(url=url,
                                           username=username,
                                           password=password,
                                           api_base_path=api_base_path,
                                           verify_ssl=verify_ssl,
                                           con_type=con_type if con_type else "CiscoAPI",
                                           DEBUG=DEBUG
                                           )
        self.headers["Accept"] = "application/json"
        self.api_user_id = api_user_id
        self.access_token = None
        self.token_type = None
        self.timestamp = None

    def _update_headers(self):
        self.headers["Authorization"] = "{} {}".format(self.token_type, self.access_token)
        self.headers["Accept"] = "application/json"
        self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}

    def _load_credentials(self):

        with open(file="{}/csapi_credentials.json".format(DATA_PATH), mode="r") as f:
            credentials = json.load(fp=f)[self.api_user_id]
            for value in credentials.values():
                if value == "":
                    return False
            self.username = credentials["username"]
            self.password = credentials["password"]
            self.access_token = credentials["access_token"]
            self.token_type = credentials["token_type"]
            self.timestamp = datetime.strptime(credentials["timestamp"], TIMESTAMP_FORMAT)
            try:
                timestamp = datetime.strptime(credentials["timestamp"], TIMESTAMP_FORMAT)
                if (datetime.now()-timestamp).total_seconds() >= 3599:
                    self.logger.debug(msg="The loaded token has expired, getting new token...")
                    self._authorize()
                else:
                    self.logger.debug(msg="The loaded token should still be valid.")
            except Exception as e:
                self.logger.error(msg="Could not verify token based on timestamp. Exception: {}".format(repr(e)))
            self._update_headers()
            self.authorized = True
            self.logger.debug(msg="Successfully loaded credentials for api_user_id='{}'".format(self.api_user_id))
            return True

    def _store_credentials(self):
        self.logger.info(msg="Storing credentials.")
        credentials = {}
        try:
            with open(file="{}/csapi_credentials.json".format(DATA_PATH), mode="r") as f:
                credentials = json.load(f)
            credentials[self.api_user_id] = {"username": self.username, "password": self.password, "access_token": self.access_token, "token_type": self.token_type, "timestamp": datetime.strftime(datetime.now(), TIMESTAMP_FORMAT)}
            with open(file="{}/csapi_credentials.json".format(DATA_PATH), mode="w") as f:
                json.dump(obj=credentials, fp=f, indent=2)
            self.logger.debug(msg="Successfully updated stored credentials for api_user_id='{}'".format(self.api_user_id))

        except FileNotFoundError:
            with open(file="{}/csapi_credentials.json".format(DATA_PATH), mode="w") as f:
                json.dump(obj={self.api_user_id: {"username": self.username, "password": self.password, "access_token": self.access_token, "token_type": self.token_type, "timestamp": datetime.strftime(datetime.now(), TIMESTAMP_FORMAT)}}, fp=f, indent=2)
                self.logger.debug(msg="Successfully stored credentials for api_user_id='{}'".format(self.api_user_id))
        except Exception as e:
            self.logger.critical(msg="Could not store credentials. Exception: '{}'".format(repr(e)))

    def _authorize(self):

        auth_url = "https://cloudsso.cisco.com/as/token.oauth2"
        payload = "client_id={}&client_secret={}&grant_type=client_credentials".format(self.username, self.password)
        response = requests.post(url=auth_url, data=payload, headers={'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'application/json'}, verify=self.verify_ssl)
        status_code = response.status_code
        if status_code == 200:
            response = response.json()
            print(response)
            self.access_token = response["access_token"]
            self.token_type = response["token_type"]
            self.timestamp = datetime.strftime(datetime.now(), TIMESTAMP_FORMAT)
            self.logger.debug(msg="Successfully obtained access token: '{}'".format(self.access_token))
            self._update_headers()
            self._store_credentials()
            self.authorized = True
        elif status_code == 401:
            try:
                if response.json()["error_description"] == "Invalid client or client credentials":
                    self.logger.critical(msg="Could not get Access Token. Reason: Invalid Credentials. Please provide valid credentials.")
                    self.authorized = False
            except Exception as e:
                self.logger.error(msg="_Authorize: Unhandled Exception occurred. Exception: {}".format(repr(e)))
                self.authorized = False
        else:
            print(status_code)
            print(response.json())

    def get(self, path, params=None):
        """
        Function providing HTTP GET method for retrieving data from APIC-EM API.

        :param path: Path of API resource, such as "/network-device". Each path must begin with "/" forward-slash character
        :param params: Dictionary of parameters for request, such as {"deviceId": "<ID of device in APIC-EM database>"}
        :return: Dictionary representation of response content under "response" key, eg. <request_response_object>show_ip_route.json()["response"] or ``None``
        """
        if not self.authorized:
            self.logger.error(msg="Connection is not authorized")
            return None
        response, status_code = self._response_handler(self._get(path=path, params=params))
        if status_code in range(200, 208):
            #print(response)
            return response
        elif status_code == 401 or status_code == 403:
            self.authorized = False
            self._authorize()
            return self.get(path=path, params=params)
        else:
            self.logger.error(msg="GET: Could not retrieve valid response for path '{}'. Response: {}".format(path, response))
            return None

    def worker_get(self, path, params, in_queue, outputs):
        while not in_queue.empty():
            chunk = in_queue.get()
            query = ",".join(chunk)
            response = self.get(path="/{}/{}".format(path, query))
            outputs.append(response)
            in_queue.task_done()


    def get_threaded(self, path, queries, max_query_len, workers, params=None):
        if isinstance(queries, str):
            queries = queries.split(",")
        # Make sure queries don't contain trailing spaces
        queries = [x.strip() for x in queries]
        # Split queries by max_query_length
        chunks = [queries[x:x + max_query_len] for x in range(0, len(queries), max_query_len)]
        if len(chunks) < workers:
            workers = len(chunks)
        in_queue = Queue()
        for chunk in chunks:
            in_queue.put(chunk)
        threads = []
        outputs = []
        for i in range(workers):
            t = Thread(name="QueryThread-{}".format(i), target=self.worker_get, args=(path, params, in_queue, outputs))
            threads.append(t)
        [t.start() for t in threads]
        in_queue.join()
        return outputs


if __name__ == '__main__':
    client = CiscoApiBase(url="https://api.cisco.com", api_user_id="an", api_base_path="/supporttools/eox/rest/5", DEBUG=True)
    client._initialize()
    client.get_threaded(path="/EOXByProductID", max_query_len=20, workers=10, queries=["WS-C3850-24T-L","WS-C3850-48T-L","WS-C3850-24P-L","WS-C3850-24U-L","WS-C3850-48P-L","WS-C3850-48F-L","WS-C3850-48U-L","WS-C3850-24T-S","WS-C3850-48T-S","WS-C3850-24P-S","WS-C3850-24U-S","WS-C3850-48P-S","WS-C3850-48F-S","WS-C3850-48U-S","WS-C3850-24T-E","WS-C3850-48T-E","WS-C3850-24P-E","WS-C3850-24U-E","WS-C3850-48P-E","WS-C3850-48F-E","WS-C3850-48U-E","WS-C3850-12X48U-L","WS-C3850-12X48U-S","WS-C3850-12X48U-E","WS-C3850-24XU-L","WS-C3850-24XU-S","WS-C3850-24XU-E","WS-C3850-12S-S","WS-C3850-12S-E","WS-C3850-24S-S","WS-C3850-24S-E","WS-C3850-12XS-S","WS-C3850-12XS-E","WS-C3850-24XS-S","WS-C3850-24XS-E","WS-C3850-48XS-S","WS-C3850-48XS-E","WS-C3850-48XS-F-S","WS-C3850-48XS-F-E","WS-C3850-24PW-S","WS-C3850-48PW-S","WS-C3850-24UW-S","WS-C3850-48W-S ","WS-C3850-48UW-S","WS-C3850-24XUW-S","WS-C3850-12X48UW-S","WS-C3850-16XS-S","WS-C3850-16XS-E","WS-C3850-32XS-S","WS-C3850-32XS-E"])


import os
import sys
import json
from nuaal.utils import get_logger, check_path
from nuaal.definitions import ROOT_DIR, DATA_PATH



class Writer:
    """
    This object handles writing structured data in JSON format in tabular formats, such as CSV or Microsoft Excel (.xlsx).
    """
    def __init__(self, type, DEBUG=False):
        """

        :param str type: String representation of writer type, used in logging messages.
        :param bool DEBUG: Enables/disables debugging output.
        """
        self.type = type
        self.logger = get_logger(name="{}-Writer".format(self.type), DEBUG=DEBUG)

    def __str__(self):
        return "[{}-Writer]".format(self.type)

    def __repr__(self):
        return "[{}-Writer]".format(self.type)

    def json_to_lists(self, data):
        """
        This function transfers list of dictionaries into two lists, one containing the column headers (keys of dictionary) and the
        other containing individual list of values (representing rows).

        :param list data: List of dictionaries with common structure
        :return: Dict with "headers" list and "list_data" list containing rows.
        """
        headers = []
        list_data = []
        if isinstance(data, list) and len(data) > 0:
            if isinstance(data[0], dict):
                headers = list(data[0].keys())
                self.logger.debug(msg="Successfully created headers: {}".format(headers))
                for entry in data:
                    entry_list = []
                    for element in [entry[x] for x in headers]:
                        if isinstance(element, list):
                            entry_list.append(str(element))
                        else:
                            entry_list.append(element)
                    list_data.append(entry_list)
                return {"headers": headers, "list_data": list_data}

    def _get_headers(self, data):
        """
        This function returns sorted list of column headers (dictionary keys).
        :param list|dict data: List or dict of dictionaries containing common keys.
        :return: List of headers.
        """
        if isinstance(data, dict):
            headers = list(data.keys())
            headers.sort()
            return headers
        elif isinstance(data, list):
            headers = set()
            for entry in data:
                if isinstance(entry, dict):
                    for key in entry.keys():
                        headers.add(key)
            headers = list(headers)
            headers.sort()
            return headers

    def combine_data(self, data):
        """
        Function for combining data from multiple sections. Each section represents data gathered by some of the `get_` functions. Each returned dataset includes
        common headers identifying device, from which the data originates.

        :param list data: List of dictionaries, content of single section
        :return: (list) section_headers, (list) section_content
        """
        sections = set()
        section_headers = {}
        section_content = {}
        common_headers = ["hostname", "ipAddress"]
        if isinstance(data, list):
            # Create sections and sections headers
            for device in data:
                print(device["hostname"], device["ipAddress"])
                for section in device.keys():
                    if section not in common_headers:
                        section_content[section] = []
                        sections.add(section)
                        if isinstance(device[section], list):
                            section_headers[section] = common_headers + list(device[section][0].keys())
                            for entry in device[section]:
                                section_content[section].append([device[x] for x in common_headers] + [entry[x] for x in section_headers[section][len(common_headers):]])
            #print(sections)
            #print(section_headers)
            #print(json.dumps(section_content, indent=1))
            return section_headers, section_content

        elif isinstance(data, str):
            self.logger.warning(msg="Given data is a string. List of dictionaries is preferred.")
            try:
                data = json.loads(data)
                return self.json_to_lists(data=data)
            except:
                self.logger.critical(msg="Given data is not a valid JSON string.")
                return None

        pass


    def combine_device_data(self, data):
        # This function might be overly complicated for its purpose
        # Currently it tries to handle not consistent data, where different devices might have
        # different section and dictionary keys...
        common_headers = ["hostname", "ipAddress"]
        sections = set()
        section_headers = {}
        section_data = {}
        if isinstance(data, list):
            for device in data:
                for section in device.keys():
                    if section not in common_headers:
                        sections.add(section)
            sections = list(sections)
            sections.sort()
            for section in sections:
                section_headers[section] = []
                section_data[section] = []
            for device in data:
                for section in sections:
                    try:
                        #print(device[section])
                        #print(self._get_headers(device[section]))
                        for header in self._get_headers(device[section]):
                            if header not in section_headers[section]:
                                section_headers[section].append(header)
                    except Exception as e:
                        self.logger.error(msg="Combine Device Data raised an Exception: {}. Maybe inconsistent data?".format(repr(e)))
            for section in section_headers.keys():
                section_headers[section] = common_headers + section_headers[section]
        else:
            self.logger.critical(msg="Given data is not a list of dictionaries.")
        for device in data:
            for section in sections:
                #print(section_headers[section])
                for entry in device[section]:
                    temp = []
                    for header in section_headers[section][:len(common_headers)]:
                        temp.append(device[header])
                    for header in section_headers[section][len(common_headers):]:
                        try:
                            temp.append(entry[header])
                        except Exception as e:
                            self.logger.warning(msg="Exception raised: {}, maybe missing key?".format(repr(e)))
                            temp.append(None)
                    section_data[section].append(temp)
        return section_headers, section_data





from nuaal.utils import get_logger
from nuaal.definitions import LOG_PATH

class IP:
    def __init__(self, ip, DEBUG=False):
        self.logger = get_logger(name="IP", logfile_path="{}/ip_log.txt".format(LOG_PATH), DEBUG=DEBUG)
        self.decAddress = None
        self.binAddress = None
        self.decMask = None
        self.binMask = None
        self.slashMask = None
        self.binNetAddress = None
        self.decNetAddress = None
        self.binBcastAddress = None
        self.decBcastAddress = None

        (addr, mask) = None, None
        if "/" in ip:
            (addr, mask) = ip.split("/")
        elif " " in ip:
            (addr, mask) = ip.split(" ")
        else:
            (addr, mask) = ip, None
        self.decAddress = addr
        self.binAddress = self.decToBin(addr)
        if mask:
            try:
                self.binMask = self.maskToBin(mask)
                self.decMask = self.binToDec(self.binMask)
                self.slashMask = self.binMask.index("0")
                self.binNetAddress = self.binAddress[:self.slashMask] + "0" * (32 - self.slashMask)
                self.binBcastAddress = self.binAddress[:self.slashMask] + "1" * (32 - self.slashMask)
                self.decNetAddress = self.binToDec(self.binNetAddress)
                self.decBcastAddress = self.binToDec(self.binBcastAddress)

            except AttributeError:
                pass

    def decToBin(self, dec):
        octects = [int(x) for x in dec.split(".")]
        # Check
        if len(octects) != 4:
            self.logger.critical(msg="Given decimal representation doesn't have correct length. Expected 4 octect, got {}".format(len(octects)))
            raise ValueError("Given decimal representation doesn't have correct length. Expected 4 octect, got {}".format(len(octects)))
        return "".join(bin(int(x) + 256)[3:] for x in dec.split("."))

    def binToDec(self, bin):
        if len(bin) != 32:
            self.logger.critical(msg="Given binary representation doesn't have correct length. Expected 24, got {}".format(len(bin)))
            raise ValueError("Given binary representation doesn't have correct length. Expected 24, got {}".format(len(bin)))
        octects = [bin[i:i + 8] for i in range(0, len(bin), 8)]
        return ".".join([str(int(x, 2)) for x in octects])

    def maskToBin(self, mask):
        # If mask is in slash notation
        try:
            mask = int(mask)
            return "1" * mask + "0" * (32 - mask)
        # If mask is in decimal notation
        except ValueError:
            mask = self.decToBin(mask)
            slashmask = mask.index("0")
            if "1" in mask[slashmask:]:
                print("Invalid netmask!")
                return None

            return mask


class IPv4Addr(IP):
    def __init__(self, address, DEBUG=False):
        super(IPv4Addr, self).__init__(ip=address, DEBUG=DEBUG)

    def __str__(self):
        return "[IPv4Addr {}/{}]".format(self.decAddress, self.slashMask)

    def __repr__(self):
        return "[IPv4Addr {}/{}]".format(self.decAddress, self.slashMask)


class IPv4Net(IP):
    def __init__(self, network, DEBUG=False):
        super(IPv4Net, self).__init__(ip=network, DEBUG=DEBUG)
        if not self.decNetAddress:
            print("No mask given!")

    def has_address(self, addr):
        if not isinstance(addr, IPv4Addr):
            self.logger.info(msg="Trying to convert {} to IPv4Addr object.".format(addr))
            try:
                addr = IPv4Addr(address=addr)
            except Exception as e:
                self.logger.critical(msg="Could not convert {} to IPv4Addr".format(type(addr)))
                raise TypeError("Could not convert {} to IPv4Addr".format(type(addr)))
        if self.binNetAddress[:self.slashMask] == addr.binAddress[:self.slashMask]:
            return True
        else:
            return False

    def has_subnet(self, addr):
        if not isinstance(addr, IPv4Net):
            self.logger.info(msg="Trying to convert {} to IPv4Addr object.".format(addr))
            try:
                addr = IPv4Net(addr)
            except Exception as e:
                self.logger.critical(msg="Could not convert {} to IPv4Net".format(type(addr)))
                raise TypeError("Could not convert {} to IPv4Net".format(type(addr)))
        if self.binNetAddress[:self.slashMask] == addr.binNetAddress[:self.slashMask]:
            return True
        else:
            return False

    def __str__(self):
        return "[IPv4Net {}/{}]".format(self.decAddress, self.slashMask)

    def __repr__(self):
        return "[IPv4Net {}/{}]".format(self.decAddress, self.slashMask)


net = IPv4Net("192.168.1.0 255.255.255.0", DEBUG=True)
print("DecAddr:", net.decAddress)
print("DecMask:", net.decMask)
print("BinAddr:", net.binAddress)
print("BinMask:", net.binMask)
print("SlashMask:", net.slashMask)
print("DecNetAddress:", net.decNetAddress)
print("binNetAddress:", net.binNetAddress)


net2 = IPv4Net("192.168.1.128 255.255.255.128", DEBUG=True)
print("DecAddr:", net2.decAddress)
print("DecMask:", net2.decMask)
print("BinAddr:", net2.binAddress)
print("BinMask:", net2.binMask)
print("SlashMask:", net2.slashMask)
print("DecNetAddress:", net2.decNetAddress)
print("binNetAddress:", net2.binNetAddress)

print(net.has_subnet(net2))
print(net2.has_subnet(net))




from nuaal.connections.api import RestBase
from nuaal.definitions import DATA_PATH
from nuaal.utils import check_path
import json
import os
import requests

class Cisco_NX_API(RestBase):
    def __init__(self, ip, username, password, verify_ssl=False, DEBUG=False):
        super(Cisco_NX_API, self).__init__(url="https://{}".format(ip),
                                           username=username,
                                           password=password,
                                           api_base_path="/ins",
                                           verify_ssl=verify_ssl,
                                           DEBUG=DEBUG,
                                           con_type="NXOS"
                                           )
    def _initialize(self):
        if self.verify_ssl == False:
            # Disable certificate warning
            try:
                requests.packages.urllib3.disable_warnings()
            except:
                self.logger.warning(msg="Failed to disable Certificate Warnings")
        self.common_headers["auth"] = (self.username, self.password)

    def craft_payload(self, commands):
        if isinstance(commands, str):
            commands = [x.strip() for x in commands.split(";")]
        commands = " ;".join(commands)
        payload = {
            "ins_api": {
                "version": "1.0",
                "type": "cli_show",
                "chunk": "0",
                "sid": "1",
                "input": commands,
                "output_format": "json"
            }
        }
        return json.dumps(payload)
    def test(self):
        self._initialize()
        payload = self.craft_payload("show cdp neighbors")
        response = self._response_handler(self._post(path="", data=payload))
        print(json.dumps(response, indent=2))


if __name__ == '__main__':
    device = Cisco_NX_API(ip="10.17.89.47",
                          username="admin",
                          password="1234Qwer",
                          DEBUG=True)
    device.test()



from nuaal.Parsers import ParserModule
import timeit
import re

class JuniperJUNOSParser(ParserModule):
    """
    Child class of `ParserModule` designed for `cisco_ios` device type.
    """
    def __init__(self, DEBUG=False):
        super(JuniperJUNOSParser, self).__init__(device_type="juniper_junos", DEBUG=DEBUG)
    


from netmiko import ConnectHandler
from netmiko.ssh_exception import *
import json
from nuaal.utils import get_logger, check_path
from nuaal.utils import Filter
from nuaal.definitions import DATA_PATH, OUTPUT_PATH
import timeit
import os

class CliBaseConnection(object):
    """
    This class represents the base object, from which other (vendor specific classes) inherit. This class is basically a wrapper class around Kirk Byers' excellent library, netmiko. Even though the netmiko library already provides pretty straightforward and easy way to access network devices, the CliBaseConnection tries to handle multiple events which can arise, such as:

    - Device is unreachable
    - Fallback to Telnet if SSH is not supported by device (and vice-versa)
    - Handles errors in outputs

    Apart from the 'send command, receive output'  this class also performs the parsing and storing outputs.
    """
    def __init__(self, ip=None, username=None, password=None, parser=None, secret=None, enable=False, store_outputs=False, DEBUG=False):
        """

        :param ip: (str) IP address or FQDN of the device you're trying to connect to
        :param username: (str) Username used for login to device
        :param password: (str) Password used for login to device
        :param parser: (ParserModule) Instance of ParserModule class which will be used for parsing of text outputs. By default, new instance of ParserModule is created.
        :param secret: (str) Enable secret for accessing Privileged EXEC Mode
        :param enable: (bool) Whether or not enable Privileged EXEC Mode on device
        :param store_outputs: (bool) Whether or not store text outputs of sent commands
        :param DEBUG: (bool) Enable debugging logging
        """
        self.ip = ip
        self.username = username
        self.password = password
        self.ssh_method = None
        self.telnet_method = None
        self.primary_method = None
        self.secondary_method = None
        self.secret = secret
        self.enable = enable
        self.provider = None
        self._get_provider()
        self.store_outputs = store_outputs
        self.enabled = False
        self.is_alive = False
        self.config = False
        self.prompt_end = [">", "#"] # The first item is for 'not-enabled mode', the second is for 'Privileged EXEC Mode'
        self.logger = get_logger(name="Connection-{}".format(self.ip), DEBUG=DEBUG)
        self.parser = parser
        self.connected = False

        self.outputs = {}
        self.data = {"ipAddress": self.ip}
        self.device = None

    def __enter__(self):
        """
        Enables usage of Python's Context Manager, using ``with`` statement.

        :return: Instance of the ``self`` object.
        """
        self._connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        This function is used when exiting Python's Context Manager.

        :return: ``None``
        """
        try:
            self.store_raw_output(self.data["hostname"], json.dumps(self.data, indent=2), ext="json")
        except KeyError:
            self.logger.error(msg="Could not store data of device {}.".format(self.ip))
        self.disconnect()

    def _get_provider(self):
        """
        Creates provider dictionary for Netmiko connection

        :return: ``None``
        """
        self.provider = {
            "ip": self.ip,
            "username": self.username,
            "password": self.password
        }
        if self.secret:
            self.provider["secret"] = self.secret

    def _connect_telnet(self):
        """
        This function tries to establish connection with device via Telnet

        :return: (``netmiko.ConnectHandler``) device
        """
        device = None
        self.provider["device_type"] = self.telnet_method
        self.logger.debug(msg="Trying to connect to device {} via Telnet...".format(self.ip))
        try:
            device = ConnectHandler(**self.provider)
        except TimeoutError:
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: TimeOut.".format(self.ip, self.telnet_method))
        except ConnectionRefusedError:
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: Connection Refused.".format(self.ip, self.telnet_method))
        # TODO: Check fix in netmiko
        except AttributeError("module 'serial' has no attribute 'EIGHTBITS'", ):
            pass
        except Exception as e:
            print(repr(e))
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: Unknown.".format(self.ip, self.telnet_method))
        finally:
            if device:
                self.logger.info(msg="Connected to '{}' using '{}'.".format(self.ip, self.telnet_method))
            return device

    def _connect_ssh(self):
        """
        This function tries to establish connection with device via SSH

        :return: (``netmiko.ConnectHandler``) device
        """
        device = None
        self.logger.debug(msg="Trying to connect to device {} via SSH...".format(self.ip))
        self.provider["device_type"] = self.ssh_method
        try:
            device = ConnectHandler(**self.provider)
        except NetMikoTimeoutException:
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: Timeout.".format(self.ip, self.ssh_method))
        except NetMikoAuthenticationException:
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: Authentication Failed.".format(self.ip, self.ssh_method))
        # TODO: Check fix in netmiko
        except AttributeError("module 'serial' has no attribute 'EIGHTBITS'",):
            pass
        except Exception as e:
            print(repr(e))
            self.logger.error(msg="Could not connect to '{}' using '{}'. Reason: Unknown.".format(self.ip, self.ssh_method))
        finally:
            if device:
                self.logger.info(msg="Connected to '{}' using '{}'.".format(self.ip, self.ssh_method))
            return device

    def _connect(self):
        """
        This function handles connection to device, if primary method fails, it will try to connect using secondary method.

        :return: ``None``
        """
        if self.device is not None:
            if self.device.is_alive():
                self.is_alive = True
                self.logger.debug(msg="Connection is already established.")
            else:
                try:
                    self.logger.debug(msg="Trying to re-establish connection to device.")
                    self.device.establish_connection()
                    self.device.session_preparation()
                    self._check_enable_level(self.device)
                except Exception as e:
                    self.logger.critical(msg="Failed to reconnect do device.")
        else:
            self.is_alive = False
            device = None

            if self.primary_method == self.ssh_method:
                device = self._connect_ssh()
            elif self.primary_method == self.telnet_method:
                device = self._connect_telnet()
            if not device:
                if self.secondary_method == self.telnet_method:
                    device = self._connect_telnet()
                elif self.secondary_method == self.ssh_method:
                    self._connect_ssh()
            if device is not None:
                self._check_enable_level(device)
            else:
                self.logger.error(msg="Could not connect to device '{}'".format(self.ip))

    def _check_enable_level(self, device):
        """
        This function is called at the end of ``self._connect()`` to ensure that the connection is actually alive
        and that the proper privilege level is set.

        :param device: (``Netmiko.ConnectHandler``) Instance of ``netmiko.ConnectHandler``. If the connection is working, this will be set as ``self.device``
        :return: ``None``
        """
        try:
            if device.is_alive():
                self.is_alive = True
            else:
                self.logger.critical(msg="Connection is not alive.")
            prompt = device.find_prompt()
            self.data["hostname"] = prompt[:-1]
            if prompt[-1] == self.prompt_end[0]:
                self.enabled = False
            if prompt[-1] == self.prompt_end[1]:
                self.enabled = True
            if self.enable and not self.enabled:
                device.enable()
                if device.find_prompt()[-1] == self.prompt_end[1]:
                    self.logger.debug(msg="Successfully enabled Privileged EXEC Mode on device '{}'".format(self.ip))
                    self.enabled = True
                else:
                    self.logger.error(msg="Failed to enable Privileged EXEC Mode on device '{}'".format(self.ip))
            if not self.enable and self.enabled:
                device.exit_enable_mode()
                if device.find_prompt()[-1] ==self.prompt_end[0]:
                    self.logger.debug(msg="Successfully disabled Privileged EXEC Mode on device '{}'".format(self.ip))
                    self.enabled = True
                else:
                    self.logger.error(msg="Failed to disable Privileged EXEC Mode on device '{}'".format(self.ip))
        except ValueError as e:
            self.logger.critical(msg="Could not enter enable mode: {}".format(repr(e)))
        except Exception as e:
            print(repr(e))
        finally:
            self.device = device

    def disconnect(self):
        """
        This function handles graceful disconnect from the device.

        :return: ``None``
        """
        if self.device is not None:
            self.device.disconnect()
            if not self.device.is_alive():
                self.is_alive = False
                self.logger.info(msg="Successfully disconnected from device {}".format(self.ip))
            else:
                self.is_alive = True
                self.logger.error(msg="Failed to disconnect from device {}".format(self.ip))
        else:
            self.logger.info(msg="Device {} is not connected.".format(self.ip))

    def _send_command(self, command, expect_string=None):
        """

        :param str command: Command to send to device
        :return: Plaintext output of command from device
        """
        if (not self.device) or (not self.is_alive):
            self.logger.error(msg="Device {} is not connected, cannot send command.".format(self.ip))
            return None
        self.logger.debug(msg="Sending command '{}' to device {} ({})".format(command, self.data["hostname"], self.ip))
        output = None

        try:
            output = self.device.send_command(command_string=command, expect_string=expect_string)
        except AttributeError:
            self.logger.critical(msg="Connection to device {} has not been initialized.".format(self.ip))
        except Exception as e:
            self.logger.error(msg="Unhandled exception occurred when trying to send command. Exception: {}".format(repr(e)))
        finally:
            return output


    def _send_commands(self, commands):
        """
        Sends multiple commands to device.

        :param list commands: List of commands to run
        :return: Dictionary with key=command, value=output_of_the_command
        """
        output = {}
        for command in commands:
            output[command] = self._send_command(command)
        return output

    def _command_handler(self, action, filter=None):
        """
        This function tries to send multiple 'types' of given command and waits for correct output.
        This should solve the problem with different command syntax, such as 'show mac address-table' vs
        'show mac-address-table' on different versions of Cisco IOS.
        When correct output is returned, it is then parsed and the result is returned.

        :param list commands: List of command string to try, such as ['show mac-address-table', 'show mac address-table']
        :return: JSON representation of command output
        """
        start_time = timeit.default_timer()
        commands = self.command_mappings[action]
        command_output = ""
        used_command = ""
        parsed_output = []
        for command in commands:
            command_output = self._send_command(command)
            if not command_output:
                self.logger.error(msg="Could not retrieve any output. Possibly non-active connection.")
                return []
            if "% Invalid input detected at '^' marker." in command_output:
                self.logger.error(msg="Device {} does not support command '{}'".format(self.ip, command))
            elif "% Ambiguous command:" in command_output:
                self.logger.error(msg="Device {}: Ambiguous command: '{}'".format(self.ip, command))
            elif command_output == "":
                self.logger.error(msg="Device {} returned empty output for command '{}'".format(self.ip, command))
            else:
                self.logger.debug(msg="Device {} returned output for command '{}'".format(self.ip, command))
                used_command = command
                break
        if self.store_outputs and command_output != "":
            self.store_raw_output(command=used_command, raw_output=command_output)
        if command_output == "" or command_output is None:
            return []
        # Try parsing the output
        try:
            parsed_output = self.parser.autoparse(command=commands[0], text=command_output)
            if isinstance(filter, Filter):
                parsed_output = filter.universal_cleanup(data=parsed_output)
            self.data[action[4:]] = parsed_output
        except Exception as e:
            print(repr(e))
            self.logger.error(msg="Device {}: Failed to parse output of command '{}'".format(self.ip, used_command))
        finally:
            self.logger.debug(msg="Processing of action {} took {} seconds.".format(action, timeit.default_timer()-start_time))
            return parsed_output

    def store_raw_output(self, command, raw_output, ext="txt"):
        """
        This function is used for storing the plaintext output of the commands called on the device in separate files. Used mainly for debugging and
        development purposes. This function is only called if the `store_outputs` parameter is set to `True`.

        :param str command: Command string executed on the device.
        :param str raw_output: Plaintext output of the command.
        :param str ext: Extension of the file, ".txt" by default.
        :return: ``None``
        """
        path = os.path.join(OUTPUT_PATH, self.ip)
        path = check_path(path)
        if path:
            with open(os.path.join(path, "{}.{}".format(command, ext)), mode="w+") as f:
                f.write(raw_output)

    def check_connection(self):
        """
        This function can be used to check state of the connection. Returns `True` if the connection is active and `False` if it isn't.

        :return: Bool value representing the connection state.
        """
        if self.device is not None:
            if self.device.is_alive():
                self.is_alive = True
                return True
            else:
                self.logger.error(msg="Connection is prepared, but not established.")
                self.is_alive = False
                return False
        else:
            self.logger.error(msg="Connection has not been initialized.")
            self.is_alive = False
            return False

    def config_mode(self):
        if self.check_connection():
            if self.device.check_config_mode():
                self.logger.debug(msg="Configuration mode is already enabled.")
                return True
            else:
                self.device.config_mode()
                if self.device.check_config_mode():
                    self.logger.info(msg="Configuration mode has been enabled.")
                    return True
                else:
                    self.logger.error(msg="Failed to enter configuration mode.")
                    return False
        else:
            self.logger.error(msg="Could not enter device configuration mode. Reason: Connection is not established.")
            return False



    #####################
    ### GET Functions ###
    #####################

    def get_vlans(self):
        """
        This function returns JSON representation of all VLANs enabled on the device, together with list of assigned interfaces. In Cisco terms, this represents
        the `show vlan brief` command.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_vlans")

    def get_inventory(self):
        """
        This function return JSON representation of all installed modules and HW parts of the device. In Cisco terms, this represents the command `show inventory`.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_inventory")

    def get_interfaces(self):
        """
        This function returns JSON representation of all physical and virtual interfaces of the device, containing all available info about each interface.
        In Cisco terms, this represents usage of command `show interfaces`.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_interfaces")

    def get_portchannels(self):
        """
        This function returns JSON representation of all logical bind interfaces (etherchannels, portchannels). In Cisco terms, this represents the
        `show etherchannel summary` command.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_portchannels")

    def get_license(self):

        """
        This function return JSON representation of licenses activated or installed on the device. In Cisco terms, this represents the `show license` command.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_license")

    def get_version(self):
        """
        Returns JSON representation of basic device information, such as vendor, device platform, software version etc. In Cisco terms, this represents the
        command `show version`.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_version")

    def get_mac_address_table(self):
        """
        Returns content of device MAC address table in JSON format. In Cisco terms, this represents the command `show mac address-table`.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_mac_address_table")

    def get_arp(self):
        """
        Returns content of device ARP table in JSON format. In Cisco terms, this represents the command `show ip arp`.

        :return: List of dictionaries.
        """
        return self._command_handler(action="get_arp")

    def __str__(self):
        return "[Connection -> {}]".format(self.ip)

    def __repr__(self):
        return "[Connection -> {}]".format(self.ip)


from nuaal.definitions import DATA_PATH
from nuaal.utils import get_logger, check_path
import os
import json
import re
import timeit

class PatternsLib:
    def __init__(self, device_type, DEBUG=False):
        start_time = timeit.default_timer()
        self.device_type = device_type
        self.logger = get_logger(name="PatternsLib_{}".format(self.device_type), DEBUG=DEBUG)
        self.compiled_patterns = {}
        self._compile_all()
        total_time = round((timeit.default_timer() - start_time) * 1000, 3)
        self.logger.debug(msg="Initialization of PatternsLib took {} ms.".format(total_time))


    def _dir_modules(self):
        path = check_path(os.path.abspath(os.path.join(DATA_PATH, "patterns", self.device_type)))
        modules = [x for x in os.listdir(path) if x[-5:] == ".json"]
        self.logger.debug(msg="Found {} pattern modules for '{}'.".format(len(modules), self.device_type))
        module_paths = {}
        for module in modules:
            module_path = os.path.abspath(os.path.join(path, module))
            module_name = module[:-5].replace("_", " ")
            module_paths[module_name] = module_path
        return module_paths

    def _compile_pattern(self, pattern_dict):
        if not isinstance(pattern_dict, dict):
            self.logger.error(msg="Given parameter 'pattern_dict' is not a dictionary.")
        # Take pattern and flags from dictionary and return compiled pattern
        try:
            return re.compile(pattern=pattern_dict["pattern"], flags=pattern_dict["flags"])
        except Exception as e:
            # TODO: More specific exceptions
            self.logger.error(msg="Encountered exception when compiling pattern '{}'. Exception: {}".format(pattern_dict["pattern"], repr(e)))

    def _compile_module(self, module_path):
        start_time = timeit.default_timer()
        with open(module_path, mode="r") as file:
            pattern_data = json.load(file)

        compiled_pattern_data = {"command": pattern_data["command"]}
        for level in [x for x in pattern_data.keys() if "level" in x]:
            if isinstance(pattern_data[level], list):
                compiled_pattern_data[level] = []
                for pattern_dict in pattern_data[level]:
                    compiled_pattern_data[level].append(self._compile_pattern(pattern_dict=pattern_dict))
            elif isinstance(pattern_data[level], dict):
                compiled_pattern_data[level] = {}
                for key, patterns in pattern_data[level].items():
                    compiled_pattern_data[level][key] = []
                    for pattern_dict in patterns:
                        compiled_pattern_data[level][key].append(self._compile_pattern(pattern_dict=pattern_dict))
        self.logger.debug(msg="Compiling of patterns for '{}' took {} ms.".format(compiled_pattern_data["command"], round((timeit.default_timer() - start_time)*1000, 3)))
        return compiled_pattern_data

    def _compile_all(self):
        start_time = timeit.default_timer()
        for command, path in self._dir_modules().items():
            self.compiled_patterns[command] = {}
            compiled_module = self._compile_module(module_path=path)
            for level in [x for x in compiled_module.keys() if "level" in x]:
                self.compiled_patterns[command][level] = compiled_module[level]
        total_time = round((timeit.default_timer() - start_time)*1000, 3)
        self.logger.debug(msg="Compiling of all patterns for '{}' took {} ms.".format(self.device_type, total_time))



from nuaal.connections.api import RestBase
from nuaal.definitions import DATA_PATH
from nuaal.utils import check_path
import json
import os
import copy
import requests
import re
import time


class ApicEmBase(RestBase):
    """
    Object providing access to Cisco's APIC-EM REST-API.
    """
    def __init__(self, url, username=None, password=None, api_base_path="/api/v1", verify_ssl=True, DEBUG=False):
        """

        :param url: URL of the APIC-EM, such as 'https://sandboxapicem.cisco.com'
        :param username: Username for authentication
        :param password: Password for authentication
        :param api_base_path: Base path for version-1 API
        :param verify_ssl: Enable/disable verification of SSL Certificate
        :param DEBUG: Enable/disable debugging output
        """
        super(ApicEmBase, self).__init__(url=url, username=username, password=password,
                                         api_base_path=api_base_path, verify_ssl=verify_ssl,
                                         DEBUG=DEBUG, con_type="APIC-EM")
        self.authorized = False
        self._initialize()

    def _authorize(self):
        """
        Function for obtaining authorization "Service Ticket" from the APIC-EM. This function is called when HTTP Status Code 401 (Unauthorized) is returned.

        :return: ``None``
        """
        if "X-Auth_Token" in self.headers.keys():
            del self.headers["X-Auth-Token"]
            self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
        response, status_code = self._response_handler(self._post(path="/ticket", data=json.dumps({"username": self.username, "password": self.password})))

        if status_code == 200:
            if "serviceTicket" in response["response"].keys():
                self.headers["X-Auth-Token"] = response["response"]["serviceTicket"]
                self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
                self.authorized = True
                self.logger.info(msg="Successfully updated Service Ticket: {}".format(self.headers['X-Auth-Token']))
                self._store_credentials()
        if status_code == 401:
            try:
                if response["response"]["errorCode"] == "INVALID_CREDENTIALS":
                    self.logger.critical(msg="Could not get Service Ticket. Reason: Invalid Credentials. Please provide valid credentials.")
                    self.authorized = False
            except Exception as e:
                self.logger.error(msg="_Authorize: Unhandled Exception occurred. Exception: {}".format(repr(e)))
                self.authorized = False

    def _load_credentials(self):
        """
        Function for loading credentials information stored in data/apicem_credentials.json file.

        :return: ``None``
        """
        with open(file="{}/apicem_credentials.json".format(DATA_PATH), mode="r") as f:
            credentials = json.load(fp=f)
            for value in credentials.values():
                if value == "":
                    return False
            self.username = credentials["username"]
            self.password = credentials["password"]
            self.headers["X-Auth-Token"] = credentials["serviceTicket"]
            self.common_headers = {"headers": self.headers, "verify": self.verify_ssl}
            self.authorized = True
            return True

    def _store_credentials(self):
        """
        Function for storing credentials information in data/apicem_credentials.json file. This function is called after successfully obtaining Service Ticket

        :return: ``None``
        """
        self.logger.info(msg="Storing credentials.")
        with open(file="{}/apicem_credentials.json".format(DATA_PATH), mode="w") as f:
            json.dump({"username": self.username, "password": self.password, "serviceTicket": self.headers["X-Auth-Token"]}, f)

    def get(self, path, params=None):
        """
        Function providing HTTP GET method for retrieving data from APIC-EM API.

        :param path: Path of API resource, such as "/network-device". Each path must begin with "/" forward-slash character
        :param params: Dictionary of parameters for request, such as {"deviceId": "<ID of device in APIC-EM database>"}
        :return: Dictionary representation of response content under "response" key, eg. <request_response_object>show_ip_route.json()["response"] or ``None``
        """
        if not self.authorized:
            self.logger.error(msg="Connection is not authorized")
            return None
        response, status_code = self._response_handler(self._get(path=path, params=params))
        if status_code in range(200, 208):
            #print(response)
            return response["response"]
        elif status_code == 401:
            self.authorized = False
            self._authorize()
            return self.get(path=path, params=params)
        else:
            self.logger.error(msg="GET: Could not retrieve valid response for path '{}'. Response: {}".format(path, response))
            return None

    def post(self, path, data=None, files=None, params=None):
        """
        Function providing HTTP POST method for sending data to APIC-EM API.

        :param str path: Path of API resource, such as "/ticket". Each path must begin with "/" forward-slash character
        :param json data: JSON string data payload
        :param dict files: Dictionary with files to upload
        :param params: Dictionary of parameters for request, such as {"deviceId": "<ID of device in APIC-EM database>"}
        :return: Dictionary representation of response content under "response" key, eg. <request_response_object>show_ip_route.json()["response"] or ``None``
        """
        if not self.authorized:
            return None
        response, status_code = self._response_handler(self._post(path=path, data=data, files=files, params=params))
        if status_code in range(200, 208):
            return response["response"]
        elif status_code == 401:
            self.authorized = False
            self._authorize()
            return self.post(path=path, data=data, files=files, params=params)
        else:
            self.logger.error(msg="GET: Could not retrieve valid response for path '{}', Response: {}".format(path, response))
            return None

    def put(self, path, data=None, files=None, params=None):
        """
        Function providing HTTP PUT method for sending data to APIC-EM API.

        :param str path: Path of API resource
        :param json data: JSON string data payload
        :param dict files: Dictionary with files to upload
        :param dict params: Dictionary of parameters for request, such as {"deviceId": "<ID of device in APIC-EM database>"}
        :return: Dictionary representation of response content under "response" key, eg. <request_response_object>show_ip_route.json()["response"] or ``None``
        """
        if not self.authorized:
            return None
        response, status_code = self._response_handler(self._put(path=path, data=data, files=files, params=params))
        if status_code in range(200, 208):
            return response["response"]
        elif status_code == 401:
            self.authorized = False
            self._authorize()
            return self.put(path=path, data=data, files=files, params=params)
        else:
            self.logger.error(msg="PUT: Could not retrieve valid response for path '{}'".format(path))
            return None

    def delete(self, path, params=None):
        """

        :param str path: Path of API resource
        :param dict params: Dictionary of parameters for request
        :return: Dictionary representation of response
        """
        if not self.authorized:
            return None
        response, status_code = self._response_handler(self._delete(path=path, params=params))
        if status_code in range(200, 208):
            return response["response"]
        elif status_code == 401:
            self.authorized = False
            self._authorize()
            return self.delete(path=path, params=params)
        else:
            self.logger.error(msg="DELETE: Could not retrieve valid response for path '{}'".format(path))
            return None

    def check_file(self, fileId=None, namespace=None, file_name=None):
        """
        Function for checking whether given file (based on it's name or ID) exists on the APIC-EM. Can be used for checking configuration or template files.

        :param str fileId: ID string of the file in APIC-EM's database
        :param str namespace: APIC-EM's namespace, under which the file should be located
        :param str file_name: Name of the file
        :return str: FileID
        """
        if fileId:
            response = self._get(path="/file/{}".format(fileId))
            status_code = response.status_code
            response = response.text
            if status_code == 200:
                self.logger.debug(msg="File with ID: '{}' exists.".format(fileId))
                return fileId
        else:
            file_id = None
            namespaces = self.get(path="/file/namespace")
            if namespace in namespaces:
                self.logger.debug(msg="File namespace {} exists.".format(namespace))
                files = self.get(path="/file/namespace/{}".format(namespace))
                for file in files:
                    try:
                        if file["name"] == file_name:
                            file_id = file["id"]
                            self.logger.info(msg="File with name {} already exists on APIC-EM with ID: '{}'".format(file_name, file_id))
                            break
                    except Exception as e:
                        self.logger.error(msg="Check File encountered unhandled Exception: {}".format(repr(e)))
                if file_id is None:
                    self.logger.info(msg="File '{}' does not exists on APIC-EM under namespace {}.".format(file_name, namespace))
                return file_id
            else:
                self.logger.info(msg="Namespace '{}' does not exists.".format(namespace))
                return file_id

    def delete_file(self, fileId=None, namespace=None, filename=None):
        """
        Function for deleting file from APIC-EM's database based on it's ID

        :param str fileId: ID string of the file in APIC-EM's database
        :param str namespace: APIC-EM's namespace, under which the file should be located
        :param str filename: Name of the file
        :return bool:
        """
        if fileId:
            response = self.get(path="/file/{}".format(fileId))
            if response:
                self.logger.debug()
            response = self.delete(path="/file/{}".format(fileId))
            if response == "File id={} is deleted successfully".format(fileId):
                return True
            else:
                return False


    def upload_file(self, namespace, file_path):
        """
        Function for uploading file to APIC-EM's database.

        :param str namespace: APIC-EM's namespace, under which the file should be uploaded
        :param str file_path: Path to the file on local machine
        :return str: FileID of the uploaded file
        """
        file_name = os.path.basename(file_path)
        file_path = check_path(os.path.dirname(file_path), create_missing=False)
        if file_path:
            file_path = os.path.abspath(os.path.join(file_path, file_name))
            if os.path.isfile(file_path):
                file_name = os.path.basename(file_path)
            else:
                self.logger.error(msg="Given file_path is not a valid path to file.")
                return None
        else:
            self.logger.error(msg="Given file_path does not exist.")
            return None
        file_id = self.check_file(namespace=namespace, file_name=file_name)
        if file_id:
            self.logger.info(msg="File '{}' in namespace '{}' already exists with ID: '{}'".format(file_name, namespace, file_id))
            return file_id
        else:
            file_path = file_path
            file = None
            if file_path:
                try:
                    file = open(file_path, mode="rb")
                    self.logger.debug(msg="Upload file: File '{}' successfully opened.".format(file_name))
                except Exception as e:
                    self.logger.error(msg="Could not open file '{}' for reading.".format(file_name))
                    return file_id
                try:
                    response = self.post(path="/file/{}".format(namespace), files={file_name: file})
                    file_id = response["id"]
                except Exception as e:
                    self.logger.error(msg="Could not upload file '{}' to namespace {}. Exception: {}".format(file_name, namespace, repr(e)))
                finally:
                    return file_id


    def task_handler(self, taskId, timeout=2, max_retries=2):
        """
        Function for handling Task results, based on taskId.
        This function periodically queries APIC-EM for the results of given task until task is either completed or timeout is reached.
        Maximum time this function waits for result is (timeout * max_retries) seconds.

        :param str taskId: ID of the task, returned by the APIC-EM after starting task.
        :param int timeout: Number of seconds to wait after each unsuccessful request
        :param int max_retries: Maximum number of requests.
        :return str: ID of the result
        """
        progress = None
        result = None
        run_count = 0

        while run_count < max_retries:
            response = self.get(path="/task/{}".format(taskId))
            print(response)
            if response is not None:
                if response["isError"]:
                    self.logger.error(msg="Task Handler: Task '{0}' failed. Progress: '{1}' failureReason: '{2}'.".format(
                        taskId,
                        response["progress"],
                        response["failureReason"]
                    )
                    )
                    return None
                progress = str(response["progress"])
                self.logger.info(msg="Task Handler : Run:{0}: Progress is: '{1}'".format(run_count, progress))
                search_result = re.search(pattern=r"^(?P<id>([A-Za-z0-9]{8}-[A-Za-z0-9]{4}-[A-Za-z0-9]{4}-[A-Za-z0-9]{4}-[A-Za-z0-9]{12}))$", flags=re.MULTILINE, string=progress)
                if search_result:
                    self.logger.debug(msg="Task Handler : Run:{0} : Found ResultID: '{1}'".format(run_count, search_result.group("id")))
                    return search_result.group("id")
                else:
                    run_count += 1
                    time.sleep(timeout)
            else:
                self.logger.error(msg="Task Handler: Failed to get response for TaskID: '{}'".format(taskId))
                return None
        self.logger.error(msg="Task Handler: Failed to get ResultID for TaskID: '{0}'. Reason: Timeout.".format(taskId))
        return None




import logging
import sys
import os
import re
import collections
from nuaal.definitions import LOG_PATH, ROOT_DIR


def get_logger(name, DEBUG=False, handle=["stderr"]):
    """
    This function provides common logging facility by creating instances of `loggers` from python standard ``logging`` library.

    :param str name: Name of the logger
    :param bool DEBUG: Enables/disables debugging output
    :param list handle: Changing value of this parameter is not recommended.
    :return: Instance of logger object
    """
    logfile_path = os.path.join(check_path(LOG_PATH), "log.txt")
    formatter = logging.Formatter("[%(asctime)s] : %(name)s - %(levelname)s - %(message)s")
    stdout_handler = logging.StreamHandler(sys.stdout)
    stderr_handler = logging.StreamHandler(sys.stderr)
    file_handler = logging.FileHandler(logfile_path)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)

    handlers = []
    if "stderr" in handle:
        handlers.append(stderr_handler)
    if "stdout" in handle:
        handlers.append(stdout_handler)


    for handler in handlers:
        handler.setFormatter(formatter)
        if DEBUG:
            handler.setLevel(logging.DEBUG)
        else:
            handler.setLevel(logging.INFO)

    root = logging.getLogger(name)
    root.setLevel(logging.DEBUG)
    root.propagate = 0
    has_handler = {"file_handler": False, "stderr_handler": False, "stdout_handler": False}
    for handler in root.handlers:
        if isinstance(handler, logging.FileHandler):
            has_handler["file_handler"] = True
        if isinstance(handler, logging.StreamHandler):
            has_handler["stderr_handler"] = True
    if not has_handler["file_handler"]:
        root.addHandler(file_handler)
    if not has_handler["stderr_handler"]:
        root.addHandler(stderr_handler)


    return root

def interface_split(interface):
    """
    Function for splitting names of interfaces to interface type and number. Example: `FastEthernet0/18` -> `FastEthernet`, `0/18`
    :param str interface: Interface name.
    :return: Interface type, interface number
    """
    int_type, int_num = None, None
    if isinstance(interface, str):
        search_result = re.search(r"(?P<int_type>\D+)(?P<int_num>\d.*)", interface)
        try:
            int_type = search_result.group("int_type")
            int_num = search_result.group("int_num")
        except:
            pass
        finally:
            return int_type, int_num

    else:
        search_result = re.search(r"interface\s(?P<int_type>\D+)(?P<int_num>\d.*)", interface.text)
        try:
            int_type = search_result.group("int_type")
            int_num = search_result.group("int_num")
        except:
            pass
        finally:
            return int_type, int_num


def vlan_range_expander(all_vlans):
    """
    Function for expanding list of allowed VLANs on trunk interface. Example: `1-4096` -> range(1, 4097). Can be used when trying to figure out whether certain
    VLAN is allowed or not. Reverse function is ``vlan_range_shortener``.

    :param all_vlans: Either list (`["1-10", "15", "20-30"]`) or string (`"1-10,15,20-30"`) of VLAN range.
    :return: List of VLAN numbers (integers).
    """
    if isinstance(all_vlans, list):
        pass
    elif isinstance(all_vlans, str):
        all_vlans = all_vlans.split(",")
    elif isinstance(all_vlans, int):
        all_vlans = [str(all_vlans)]
    full_list = []
    for vlan in all_vlans:
        if "-" in vlan:
            temp = vlan.split("-")
            full_list = full_list + list(range(int(temp[0]), int(temp[1])+1))
        else:
            full_list.append(int(vlan))
    return full_list


def vlan_range_shortener(full_vlan_list):
    """
    Function for shortening list of allowed VLANs on trunk interface. Example: `[1,2,3,4,5]` -> `["1-5"]`. Reverse function is ``vlan_range_expander``.

    :param full_vlan_list: List of integers representing VLAN numbers.
    :return: List of strings.
    """
    # TODO: Make more reliable
    shortened_vlan_list = []
    # Initialize with first element
    temp_element = str(full_vlan_list[0])

    for i in range(1, len(full_vlan_list)-1):
        if (full_vlan_list[i-1] + 1 == full_vlan_list[i]) and (full_vlan_list[i] + 1 == full_vlan_list[i+1]):
            # Do nothing
            continue
        if full_vlan_list[i-1] + 1 == full_vlan_list[i] and (full_vlan_list[i] + 1 != full_vlan_list[i+1]):
            # Close temp element and write
            temp_element = temp_element + "-" + str(full_vlan_list[i])
            shortened_vlan_list.append(temp_element)
            temp_element = ""
            continue
        if full_vlan_list[i - 1] + 1 != full_vlan_list[i] and (full_vlan_list[i] + 1 == full_vlan_list[i + 1]):
            # Begin new temp element
            temp_element = str(full_vlan_list[i])
            continue
        if full_vlan_list[i - 1] + 1 != full_vlan_list[i] and (full_vlan_list[i] + 1 != full_vlan_list[i + 1]):
            if temp_element != "":
                shortened_vlan_list.append(temp_element)
                temp_element = ""
            shortened_vlan_list.append(str(full_vlan_list[i]))
            continue
    # Close with end element
    if full_vlan_list[-1] == full_vlan_list[-2] + 1:
        temp_element = temp_element + "-" + str(full_vlan_list[-1])
        shortened_vlan_list.append(temp_element)
        temp_element = ""
    else:
        shortened_vlan_list.append(str(full_vlan_list[-1]))
    return shortened_vlan_list


def int_name_convert(int_name, out_type=None):
    """
    Function for converting long interface names to short and vice versa. Example: `"GigabitEthernet0/12"` -> `"Gi0/12"` and `"Gi0/12"` -> `"GigabitEthernet0/12"`

    :param str int_name: Interface name
    :param str out_type: Desired form of interface name, either `"long"` or `"short"`. If None, return opposite form
    :return: Interface name
    """
    int_type, int_num = interface_split(int_name)
    short = ["Eth", "Et", "Se", "Fa", "Gi", "Te", "Po", "Tu"]
    long = ["Ethernet", "Ethernet", "Serial", "FastEthernet", "GigabitEthernet", "TenGigabitEthernet", "Port-channel", "Tunnel"]
    if out_type is None:
        # Return opposite version
        if int_type in short:
            return long[short.index(int_type)] + int_num
        elif int_type in long:
            return short[long.index(int_type)] + int_num
        else:
            # Error
            return int_name
    elif out_type == "long":
        if int_type in short:
            return long[short.index(int_type)] + int_num
        elif int_type in long:
            return int_name
        else:
            # Error
            return int_name
    elif out_type == "short":
        if int_type in short:
            return int_name
        elif int_type in long:
            return short[long.index(int_type)] + int_num
        else:
            # Error
            return int_name
    else:
        return int_name


def mac_addr_convert(mac_address=u""):
    """
    Function for providing single format for MAC addresses.

    :param str mac_address: MAC address to be converted.
    :return: MAC address in format `XX:XX:XX:XX:XX:XX`.
    """
    try:
        mac = mac_address.replace(".", "")
        mac = [mac[i:i + 2].upper() for i in range(0, len(mac), 2)]
        mac = ":".join(mac)
        return mac
    except Exception as e:
        return mac_address


def update_dict(orig_dict, update_dict):
    """
    Function for updating dictionary values.

    :param dict orig_dict: Dictionary to be updated.
    :param dict update_dict: Dictionary to update from.
    :return: Updated dictionary.
    """
    for k, v in update_dict.items():
        if isinstance(v, collections.Mapping):
            orig_dict[k] = update_dict(orig_dict.get(k, {}), v)
        else:
            orig_dict[k] = v
    return orig_dict


def check_path(path, create_missing=True):
    """
    Function for checking path availability, returns `path` if specified folder exists, ``False`` otherwise. Can also create missing folders. Used for handling
    absolute and relative paths.

    :param str path: Path of the folder.
    :param bool create_missing: Whether or not to create missing folders.
    :return: Bool
    """
    if os.path.isabs(path):
        if not os.path.exists(path):
            if create_missing:
                os.makedirs(path)
                path = os.path.abspath(path)
                return path
            else:
                return False
        else:
            return os.path.abspath(path)
    else:
        # Prepend ROOT_DIR for relative paths
        path = os.path.abspath(os.path.join(ROOT_DIR, path))
        if not os.path.exists(path):
            if create_missing:
                os.makedirs(path)
                return path
            else:
                return False
        else:
            return path
